#!/usr/bin/python3
"""
    The MIT License (MIT)

    Copyright (c) 2023 pkjmesra

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

"""
import json
from time import sleep
from collections import namedtuple
from mthrottle import Throttle
from PKNSETools.Benny import NSE
import requests
from PKDevTools.classes import Archiver

from PKNSETools.PKConstants import (_autoComplete_url_path, _base_domain,
                                    _head, _quote_url_path)
from PKDevTools.classes.OutputControls import OutputControls

session = requests.session()

throttleConfig = {
    'default': {
        'rps': 3,
    },
}
MAX_PENALTY_COUNT = 1
MIN_PENALTY_WAIT_SECONDS = 10
th = Throttle(throttleConfig, MAX_PENALTY_COUNT)

def initialize():
    session = NSE(download_folder=Archiver.get_user_cookies_dir()).session
    
def _get_Tuple_From_JSON(companyDict):
    return namedtuple('X', companyDict.keys())(*companyDict.values())

def get_Id_By_Name(name):
    companyDetails = get_Company_Details_By_Name(name)
    return companyDetails.info.identifier

def get_Search_Results_By_Name(name):
    search_url = f'{_base_domain}{_autoComplete_url_path}'
    session.get(f'{_base_domain}', headers=_head)
    search_results = session.get(url=search_url.format(name), headers=_head)
    # {
    # "symbols":[
    #     {
    #         "symbol":"SBIN",
    #         "symbol_info":"State Bank of India",
    #         "symbol_suggest":[
    #             {
    #             "input":"SBIN",
    #             "weight":16
    #             },
    #             {
    #             "input":"State",
    #             "weight":8
    #             },
    #             {
    #             "input":"Bank",
    #             "weight":8
    #             },
    #             {
    #             "input":"of",
    #             "weight":8
    #             },
    #             {
    #             "input":"India",
    #             "weight":8
    #             }
    #         ],
    #         "result_type":"symbol",
    #         "result_sub_type":"equity",
    #         "activeSeries":[
    #             "EQ"
    #         ],
    #         "url":"/get-quotes/equity?symbol=SBIN"
    #     }
    # ],
    # "mfsymbols":[
        
    # ],
    # "search_content":[
    #     {
    #         "title":"अंडरलाइंग्स एंड इन्फ़ॉर्मेशन की लिस्ट",
    #         "content":"LIMITEDRELIANCE160SBI CARDS AND PAYMENT SERVICES LIMITEDSBICARD161SBI LIFE INSURANCE COMPANY\nLIMITEDSBILIFE162SHREE CEMENT LIMITEDSHREECEM163SHRIRAM TRANSPORT FINANCE COMPANY LIMITEDSRTRANSFIN164SIEMENS LIMITED\nSIEMENS165SRF LIMITEDSRF166STATE BANK OF <b>INDIASBIN167STEEL</b>",
    #         "url":"/products-services/equity-derivatives-list-underlyings-information-hindi",
    #         "result_type":"content",
    #         "result_sub_type":"content"
    #     },
    #     {
    #         "title":"List of Underlyings and Information",
    #         "content":"LIMITEDRECLTD150RELIANCE INDUSTRIES LIMITEDRELIANCE151SBI CARDS AND PAYMENT SERVICES LIMITEDSBICARD\n152SBI LIFE INSURANCE COMPANY LIMITEDSBILIFE153SHREE CEMENT LIMITEDSHREECEM154SHRIRAM FINANCE LIMITEDSHRIRAMFIN155\nSIEMENS LIMITEDSIEMENS156SRF LIMITEDSRF157STATE BANK OF <b>INDIASBIN158STEEL</b>",
    #         "url":"/products-services/equity-derivatives-list-underlyings-information",
    #         "result_type":"content",
    #         "result_sub_type":"content"
    #     },
    #     {
    #         "title":"અંડરલાયિંગ એન્ડ ઇન્ફોર્મેશન નું લિસ્ટ",
    #         "content":"INTERNATIONAL LIMITEDMOTHERSON151SBI CARDS AND PAYMENT SERVICES LIMITEDSBICARD152SBI LIFE INSURANCE COMPANY\nLIMITEDSBILIFE153SHREE CEMENT LIMITEDSHREECEM154SHRIRAM FINANCE LIMITEDSHRIRAMFIN155SIEMENS LIMITEDSIEMENS156SRF LIMITED\nSRF157STATE BANK OF <b>INDIASBIN158STEEL</b>",
    #         "url":"/products-services/equity-derivatives-list-underlyings-information-gujarati",
    #         "result_type":"content",
    #         "result_sub_type":"content"
    #     },
    #     {
    #         "title":"Press Releases - Archives",
    #         "content":"Security listed & admitted to dealings � GRAVITA\n * NSE completes its 2680th Normal Settlement\n\nNov 12, 2010\n * Market-wide Position Limit in NAGARFERT\n * NSE completes its 2679th Normal Settlement\n\nNov 11, 2010\n * Security listed & admitted to dealings � <b>SBIN</b>",
    #         "url":"/resources/exchange-communication-press-releases-archives",
    #         "result_type":"content",
    #         "result_sub_type":"content"
    #     }
    # ],
    # "sitemap":[
        
    # ]
    # }  
    search_result = json.loads(str(search_results.json()).replace("'","\""), object_hook=_get_Tuple_From_JSON)
    return search_result

def download(symbolOrTask, trialCount=0):
    get_details = f'{_base_domain}{_quote_url_path}'
    symbol = symbolOrTask
    if not isinstance(symbolOrTask, str):
        symbol = symbolOrTask.userData    
    company_details = NSE(download_folder=Archiver.get_user_cookies_dir()).session.get(url=get_details.format(symbol), headers=_head)
    if company_details.status_code == 401 and trialCount <=2:
        return download(symbolOrTask,trialCount=trialCount+1)
        
    if company_details.status_code == 429 or company_details.status_code == 403:
        OutputControls().printOutput(f"{company_details.status_code}: {company_details.text}")
        if (th.penalize()):
            sleep(MIN_PENALTY_WAIT_SECONDS)
            th.maxPenaltyCount += MAX_PENALTY_COUNT
    if company_details.status_code == 200:
        # try:
        #     # Let's try without any conversions first
        #     companyDetails = json.loads(company_details.text, object_hook=_get_Tuple_From_JSON)
        # except:
        text = str(company_details.json()).replace("'","\"").replace('True','true').replace('False','false').replace('None','null')
        companyDetails = json.loads(text, object_hook=_get_Tuple_From_JSON)
            # pass
    # '{"info": {
    #       "symbol": "SBIN", 
    #       "companyName": "State Bank of India", 
    #       "industry": "BANKS", 
    #       "activeSeries": ["EQ"], 
    #       "debtSeries": [], 
    #       "tempSuspendedSeries": ["IL", "N1", "N2", "N3", "N4", "N5", "N6"], 
    #       "isFNOSec": true, 
    #       "isCASec": false, 
    #       "isSLBSec": true, 
    #       "isDebtSec": false, 
    #       "isSuspended": false, 
    #       "isETFSec": false, 
    #       "isDelisted": false, 
    #       "isin": "INE062A08066", 
    #       "isTop10": false, 
    #       "identifier": "SBINEQN"
    #       }, 
    # "metadata": { "series": "EQ", 
    #               "symbol": "SBIN", 
    #               "isin": "INE062A01020", 
    #               "status": "Listed", 
    #               "listingDate": "01-Mar-1995", 
    #               "industry": "Public Sector Bank", 
    #               "lastUpdateTime": "11-Aug-2023 16:00:00", 
    #               "pdSectorPe": 16.6, 
    #               "pdSymbolPe": 8.39, 
    #               "pdSectorInd": "NIFTY BANK                                        "}, 
    #               "securityInfo": {
    #                   "boardStatus": "Main", 
    #                   "tradingStatus": "Active", 
    #                   "tradingSegment": "Normal Market", 
    #                   "sessionNo": "-", 
    #                   "slb": "Yes", 
    #                   "classOfShare": "Equity", 
    #                   "derivatives": "Yes", 
    #                   "surveillance": {"surv": null, "desc": null}, 
    #                   "faceValue": 1, 
    #                   "issuedSize": 8924611934}, 
    #               "sddDetails": {
    #                   "SDDAuditor": "-", 
    #                   "SDDStatus": "-"}, 
    #               "priceInfo": {
    #                   "lastPrice": 574.5, 
    #                   "change": 0.8500000000000227, 
    #                   "pChange": 0.14817397367733334, 
    #                   "previousClose": 573.65, 
    #                   "open": 575.45, 
    #                   "close": 574.15, 
    #                   "vwap": 575.36, 
    #                   "lowerCP": "516.30", 
    #                   "upperCP": "631.00", 
    #                   "pPriceBand": "No Band", 
    #                   "basePrice": 573.65, 
    #                   "intraDayHighLow": {"min": 572, "max": 579, "value": 574.5}, 
    #                   "weekHighLow": {"min": 499.35, "minDate": "01-Feb-2023", "max": 629.55, "maxDate": "15-Dec-2022", "value": 574.5}, 
    #                   "iNavValue": null, 
    #                   "checkINAV": false}, 
    #               "industryInfo": {"macro": "Financial Services", "sector": "Financial Services", "industry": "Banks", 
    #                   "basicIndustry": "Public Sector Bank"}, 
    #               "preOpenMarket": {"preopen": [{"price": 516.3, "buyQty": 0, "sellQty": 104},{"price": 544.95, "buyQty": 0, "sellQty": 1554}, {"price": 550, "buyQty": 0, "sellQty": 144}, {"price": 553.1, "buyQty": 0, "sellQty": 1}, {"price": 575.45, "buyQty": 0, "sellQty": 0, "iep": true}, {"price": 599.85, "buyQty": 6, "sellQty": 0}, {"price": 600, "buyQty": 5, "sellQty": 0}, {"price": 602.35, "buyQty": 654, "sellQty": 0}, {"price": 620, "buyQty": 101, "sellQty": 0}], 
    #                   "ato": {"buy": 10667, "sell": 1621}, 
    #                   "IEP": 575.45, 
    #                   "totalTradedVolume": 29301, 
    #                   "finalPrice": 575.45, 
    #                   "finalQuantity": 29301, 
    #                   "lastUpdateTime": "11-Aug-2023 09:07:49", 
    #                   "totalBuyQuantity": 286624, 
    #                   "totalSellQuantity": 603729, 
    #                   "atoBuyQty": 10667, 
    #                   "atoSellQty": 1621}
    # }'
    if not isinstance(symbolOrTask, str):
        result = company_details.text
        if symbolOrTask.taskId > 0:
            symbolOrTask.progressStatusDict[symbolOrTask.taskId] = {'progress': 0, 'total': 1}
            symbolOrTask.resultsDict[symbolOrTask.taskId] = result
            symbolOrTask.progressStatusDict[symbolOrTask.taskId] = {'progress': 1, 'total': 1}
        else:
            symbolOrTask.result = result
    return companyDetails

def get_Company_Details_By_Name(name):
    search_results = get_Search_Results_By_Name(name)
    symbol = search_results.symbols[0].symbol
    return download(symbol)



"""
    The MIT License (MIT)

    Copyright (c) 2023 pkjmesra

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

"""
from PKNSETools.morningstartools.security import Security, DUMMY_SECURITY
import pandas as pd
from PKDevTools.classes.CookieHelper import CookieHelper
from PKDevTools.classes import Archiver
from PKDevTools.classes.Fetcher import session
from PKDevTools.classes.log import default_logger
from PKNSETools.morningstartools.NSEStockMFIDB import NSEStockMFIDB
from PKNSETools.morningstartools.NSEStockFairValueDB import NSEStockFairValueDB
from PKDevTools.classes.PKDateUtilities import PKDateUtilities

class Stock(Security):
    """
    Main class to access data about stocks, inherit from Security class
    Args:
        term (str): text to find a stock, can be a name, part of a name or the isin of the stocks
        pageSize (int): number of stocks to return
        itemRange (int) : index of stocks to return (must be inferior to PageSize)
        proxies = (dict) : set the proxy if needed , example : {"http": "http://host:port","https": "https://host:port"}

    Examples:
        >>> Stocks('0P0000712R', 9, 0)
        >>> Stocks('visa')

    Raises:
        TypeError: raised whenever the parameter type is not the type expected
        ValueError : raised whenever the parameter is not valid or no stock found

    """

    def __init__(self, term = None, exchange: str = "INDIA", pageSize : int =1, itemRange: int = 0, filters: dict = {}, proxies: dict = {}):
        
        super().__init__(term=term,asset_type='stock',exchange=exchange,pageSize=pageSize,itemRange=itemRange,filters=filters,proxies=proxies)
        self.cookieHelper = None
        self.defaultHeaders = Stock.morningstarDefaultHeaders()
        self.defaultParams = Stock.morningstarDefaultParams()
        self.defaultCookies = None
        self.refreshMorningstarTokens(self.defaultParams, self.defaultHeaders)

    @staticmethod
    def morningstarDefaultParams():
        params = {
            "secExchangeList": "",
            "locale" :"en",
            "languageId": "en",
            "clientId":"RSIN_SAL",
            "version":"4.13.0",
            "access_token":"OAg7sEUU8PQ1V8pyXApDCAyzuxqr",
            }
        return params
    
    @staticmethod
    def morningstarDefaultHeaders():
        headers = {
            "X-Api-Realtime-E": "ew0KICAiYWxnIjogIlJTQS1PQUVQIiwNCiAgImVuYyI6ICJBMTI4R0NNIg0KfQ.W4pGZbjPWyUkcwcaGY39TU2293P-E7Sy_LYI_8-h-HFUFEDSbM1Li7c4D880QOw-jQPxXdOirJIMn1bj6px4kd3Bsij5jal16cFJjtwLbCdZ_ONVUywlZLhfKHeOfaXlZCQrN6HAi9l49-iYHWAfryiAVrYTm5WIseP1inRkopg.1fPJ5lD9xM0wnRSM.8NzTksDBHjQpQMFIbImiQz6HAVLhqNNIm8rKFQ3LojEEOyB8Q9vTbMQy3bABMRv4rnahT1uI51cJlg9Fv2rilCUsBskNTDRkAh02TVHCE4z7PKOCWWWdzlNt-Yh195MhQq1Vras4_PCmkCBdEKkY8ooF5-D8kEUWl_LXy6dUj4ytvy7wGMLxMp6pgWAHMLBPW9PxAlKGWZAHs0x7qm1TIlF4M5WAdKBE4d69hyaMFqIEoNmqjnx_1sMdnTNgVK2KBgHdn007.pW2CchSpaWFh95YttagfLg",
            "X-Sal-Contenttype": "nNsGdN3REOnPMlKDShOYjlk6VYiEVLSdpfpXAm7o2Tk="
        }
        return headers

    def needsNetworkRequest(self, dbClass:None):
        # Let's check when the last network call was made. If it was today, there's no point going again!
        latestCheckedOnDate = dbClass.searchCache(ticker=DUMMY_SECURITY)
        tradingDate = PKDateUtilities.tradingDate().strftime("%Y-%m-%d")
        if latestCheckedOnDate is not None and len(latestCheckedOnDate) > 0 and DUMMY_SECURITY in latestCheckedOnDate.keys():
            latestDate = latestCheckedOnDate[DUMMY_SECURITY]
            if latestDate == tradingDate:
                return False
        if self.term == DUMMY_SECURITY or self.ticker == DUMMY_SECURITY:
            # The caller wants us to save the LatestCheckedOnDate. This should be done at the end after 
            # all the rest of the stock searches has been made.
            dbClass.saveCache(ticker=DUMMY_SECURITY, stockDict={DUMMY_SECURITY:tradingDate})
            return False
        return True
    
    def analysisData(self):
        """
        This function retrieves general data about the stock.

        Returns:
            dict with general data about stock

        Examples:
            >>> Stock("visa", exchange="nyse").analysisData()

        """
        return self.GetData("morningstarTake/v3",url_suffixe='analysisData')
    
    def analysisReport(self):
        """
        This function retrieves the analysis of the stock.

        Returns:
            dict with analyst overview

        Examples:
            >>> Stock("visa", exchange="nyse").analysisReport()

        """
        return self.GetData("morningstarTake/v4",url_suffixe='analysisReport')
    
    def balanceSheet(self, period='annual', reportType='original'):
        """
        This function retrieves the balance sheet.

        Args:
            period (str) : possible values are annual, quarterly
            reportType (str) : possible values are original, restated

        Returns:
            dict with balance sheet

        Examples:
            >>> Stock("visa", exchange="nyse").balanceSheet('quarterly', 'original')
            >>> Stock("visa", exchange="nyse").balanceSheet('annual', 'restated')

        """

        return self.financialStatement("balancesheet",period=period, reportType=reportType)
    
    def boardOfDirectors(self):
        """
        This function retrieves information about the board of directors.

        Returns:
            dict with board of directors information

        Examples:
            >>> Stock("Alphabet Inc Class A").boardOfDirectors()
    
        """
        return self.GetData("insiders/boardOfDirectors")

    def cashFlow(self, period='annual', reportType='original'):
        """
        This function retrieves the cash flow.

        Args:
            period (str) : possible values are annual, quarterly
            reportType (str) : possible values are original, restated

        Returns:
            dict with cash flow

        Examples:
            >>> Stock("visa", exchange="nyse").cashFlow('annual', 'restated')
            >>> Stock("visa", exchange="nyse").cashFlow('quarterly', 'restated')

        """

        return self.financialStatement("cashflow",period=period, reportType=reportType)
    
    def dividends(self):
        """
        This function retrieves the dividends of the stock.

        Returns:
            dict with dividends

        Examples:
            >>> Stock("visa", exchange="nyse").dividends()

        """
        return self.GetData("dividends/v4")
    
    def esgRisk(self):
        """
        This function retrieves the esg risk of the stock.

        Returns:
            dict with esg risk

        Examples:
            >>> Stock("visa", exchange="nyse").esgRisk()

        """
        return self.GetData("esgRisk")
    
    def financialHealth(self):
        """
        This function retrieves the financial health of the stock.

        Returns:
            dict with financial health

        Examples:
            >>> Stock("visa", exchange="nyse").financialHealth()

        """
        return self.GetData("keyStats/financialHealth", url_suffixe='')
    
    def financialStatement(self,statement ='summary', period='annual', reportType='original'):
        """
        This function retrieves the financial statement.

        Args:
            statement (str) : possible values are balancesheet, cashflow, incomestatement, summary
            period (str) : possible values are annual, quarterly
            reportType (str) : possible values are original, restated

        Returns:
            dict with financial statement 

        Examples:
            >>> Stock("visa", exchange="nyse").financialStatement('summary', 'quarterly', 'original')
            >>> Stock("visa", exchange="nyse").financialStatement('cashflow', 'annual', 'restated')
            >>> Stock("visa", exchange="nyse").financialStatement('balancesheet', 'annual', 'restated')
            >>> Stock("visa", exchange="nyse").financialStatement('incomestatement', 'annual', 'restated')

        """

        if not isinstance(statement, str):
            raise TypeError('statement parameter should be a string')
        
        if not isinstance(period, str):
            raise TypeError('period parameter should be a string')
        
        if not isinstance(reportType, str):
            raise TypeError('reportType parameter should be a string')
        
        statement_choice = {"balancesheet" : "balanceSheet", "cashflow" : "cashFlow", "incomestatement" : "incomeStatement", "summary" : "summary"}
        
        period_choice = {"annual" : "A", "quarterly" : "Q"}

        reportType_choice = {"original" : "A", "restated" : "R"}

        if statement not in statement_choice:
            raise ValueError(f"statement parameter must take one of the following value : { ', '.join(statement_choice)}")

        if period not in period_choice:
            raise ValueError(f"period parameter must take one of the following value : { ', '.join(period_choice)}")

        if reportType not in reportType_choice:
            raise ValueError(f"reportType parameter must take one of the following value : { ', '.join(reportType_choice.keys())}")
        
        params = {"reportType" : reportType_choice[reportType]}
        if statement == "summary":
            return self.GetData("newfinancials",params=params, url_suffixe=f"{period}/summary")
        
        params["dataType"] = period_choice[period]

        return self.GetData("newfinancials",params=params, url_suffixe=f"{statement_choice[statement]}/detail")
    
    def financialSummary(self, period='annual', reportType='original'):
        """
        This function retrieves the financial statement summary.

        Args:
            period (str) : possible values are annual, quarterly
            reportType (str) : possible values are original, restated

        Returns:
            dict with financial statement summary

        Examples:
            >>> Stock("visa", exchange="nyse").financialSummary('quarterly', 'original')
            >>> Stock("visa", exchange="nyse").financialSummary('annual', 'restated')

        """

        return self.financialStatement("summary",period=period, reportType=reportType)
    
    def freeCashFlow(self):
        """
        This function retrieves the free cash flow.

        Returns:
            dict with free cash flow

        Examples:
            >>> Stock("visa", exchange="nyse").freeCashFlow()
  
        """
        return self.GetData("keyStats/cashFlow", url_suffixe='')
    
    def historical(self, start_date,end_date,frequency="daily"):
        """
        This function retrieves the historical price, volume and divide of the stock.

        Args:
            start_date (datetime) : start date to get history
            end_date (datetime) : end date to get history
            frequency (str) : can be daily, weekly, monthly

        Returns:
            list of dict with price, volume and dividend

        Examples:
            >>> Stock("visa", exchange="nyse").history(datetime.datetime.today()- datetime.timedelta(30),datetime.datetime.today())

        """
        return self.TimeSeries(["open","high","low","close",
                                "volume","previousClose","dividend"],
                                start_date=start_date,end_date=end_date,frequency=frequency)
    def incomeStatement(self, period='annual', reportType='original'):
        """
        This function retrieves the income statement.

        Args:
            period (str) : possible values are annual, quarterly
            reportType (str) : possible values are original, restated

        Returns:
            dict with income statement

        Examples:
            >>> Stock("visa", exchange="nyse").incomeStatement('quarterly', 'original')
            >>> Stock("visa", exchange="nyse").incomeStatement('annual', 'restated')

        """
        
        return self.financialStatement("incomestatement",period=period, reportType=reportType)
    
    def institutionBuyers(self, top=50):
        """
        This function retrieves the institutions which buy the stock.

        Args:
            top (int) : number of institutions to return
        Returns:
            dict with the buyers

        Examples:
            >>> Stock("visa", exchange="nyse").institutionBuyers(top=50)
  
        """
        
        if not isinstance(top, int):
            raise TypeError('top parameter should be an integer')
        
        return self.GetData("ownership/v1", url_suffixe= f"Buyers/institution/{top}/data")
    
    def institutionConcentratedOwners(self, top=50):
        """
        This function retrieves the institutions which are concentrated on the stock.

        Args:
            top (int) : number of institutions to return
        Returns:
            dict with the concentarted owners

        Examples:
            >>> Stock("visa", exchange="nyse").institutionConcentratedOwners(top=50)

        """
        
        if not isinstance(top, int):
            raise TypeError('top parameter should be an integer')
        
        return self.GetData("ownership/v1", url_suffixe= f"ConcentratedOwners/institution/{top}/data")
    
    def institutionOwnership(self, top=5):
        """
        This function retrieves the main institutions which own the stock.

        Args:
            top (int) : number of institutions to return
        Returns:
            dict with the main owners

        Examples:
            >>> Stock("visa", exchange="nyse").institutionOwnership(top=50)
   
        """
        
        if not isinstance(top, int):
            raise TypeError('top parameter should be an integer')
        r_saved = None
        r = NSEStockMFIDB().searchCache(ticker=self.ticker)
        if r is not None and len(r) > 0 and "FII" in r.keys():
            r_saved = r["FII"]
            # Give the opportunity for the first 2 days to fetch latest end-of-last-month MFI data
            if PKDateUtilities.currentDateTime().day not in [1,2]:
                return r_saved
        elif PKDateUtilities.isTradingTime():
            # Let's not make new requests to update the values during trading hours.
            return None
        elif not self.needsNetworkRequest(NSEStockMFIDB()):
            return None
        
        params = {"component":"sal-ownership"}
        params = self.defaultParams | params
        try:
            r = self.GetData("ownership/v1", url_suffixe= f"OwnershipData/institution/{top}/data", params=params, headers=self.defaultHeaders)
        except ConnectionError as e:
            default_logger().debug(f"{e}:\n{self.term}:Going to retrieve/reset the cookies and meta tags!",exc_info=True)
            self.refreshMorningstarTokens(self.defaultParams, self.defaultHeaders)
            params = {"component":"sal-ownership"}
            params = self.defaultParams | params
            r = self.GetData("ownership/v1", url_suffixe= f"OwnershipData/institution/{top}/data", params=self.defaultParams, headers=self.defaultHeaders)
        if r is not None and len(r) > 0 and len(r["rows"]) > 0:
            NSEStockMFIDB().saveCache(ticker=self.ticker, stockDict={"FII":r})
            r_saved = r
        return r_saved
    
    def institutionSellers(self, top=50):
        """
        This function retrieves the institutions which sell on the stock.

        Args:
            top (int) : number of institutions to return
        Returns:
            dict with sellers

        Examples:
            >>> Stock("visa", exchange="nyse").institutionSellers(top=50)
  
        """
        if not isinstance(top, int):
            raise TypeError('top parameter should be an integer')
        
        return self.GetData("ownership/v1", url_suffixe= f"Sellers/institution/{top}/data")
    
    def keyExecutives(self):
        """
        This function retrieves information oabout key excutives of the company.

        Returns:
            dict with key executives information

        Examples:
            >>> Stock("visa", exchange="nyse").keyExecutives()

        """
        return self.GetData("insiders/keyExecutives")
    def keyRatio(self):
        """
        This function retrieves the key ratio of the stock.

        Returns:
            dict with key ratio

        Examples:
            >>> Stock("visa", exchange="nyse").keyRatio()

        """
        return self.GetData("keyratios")
    
    def mutualFundBuyers(self, top=50):
        """
        This function retrieves the mutual funds which buy the stock.

        Args:
            top (int) : number of mutual funds to return
        Returns:
            dict with the buyers

        Examples:
            >>> Stock("visa", exchange="nyse").mutualFundBuyers(top=50)

        """
        
        if not isinstance(top, int):
            raise TypeError('top parameter should be an integer')
        
        return self.GetData("ownership/v1", url_suffixe= f"Buyers/mutualfund/{top}/data")
    
    def mutualFundConcentratedOwners(self, top=50):
        """
        This function retrieves the mutual funds which are concentrated on the stock.

        Args:
            top (int) : number of mutual funds to return
        Returns:
            dict with the concentarted owners

        Examples:
            >>> Stock("visa", exchange="nyse").mutualFundConcentratedOwners(top=50)
 
        """
        
        if not isinstance(top, int):
            raise TypeError('top parameter should be an integer')
        
        return self.GetData("ownership/v1", url_suffixe= f"ConcentratedOwners/mutualfund/{top}/data")
    
    def mutualFundOwnership(self, top=5):
        """
        This function retrieves the main mutual funds which own the stock.

        Args:
            top (int) : number of mutual funds to return
        Returns:
            dict with the main owners

        Examples:
            >>> Stock("visa", exchange="nyse").mutualFundOwnership(top=50)
  
        """
        
        if not isinstance(top, int):
            raise TypeError('top parameter should be an integer')
        r_saved = None
        r = NSEStockMFIDB().searchCache(ticker=self.ticker)
        if r is not None and len(r) > 0 and "MF" in r.keys():
            r_saved = r["MF"]
            # Give the opportunity for the first 2 days to fetch latest end-of-last-month MFI data
            if PKDateUtilities.currentDateTime().day not in [1,2]:
                return r_saved
        elif PKDateUtilities.isTradingTime():
            # Let's not make new requests to update the values during trading hours.
            return None
        elif not self.needsNetworkRequest(NSEStockMFIDB()):
            return None
        
        params = {"component":"sal-ownership"}
        params = self.defaultParams | params
        try:
            r = self.GetData("ownership/v1", url_suffixe= f"OwnershipData/mutualfund/{top}/data", params=params, headers=self.defaultHeaders)
        except ConnectionError as e:
            default_logger().debug(f"{e}:\n{self.term}:Going to retrieve/reset the cookies and meta tags!",exc_info=True)
            self.refreshMorningstarTokens(self.defaultParams, self.defaultHeaders)
            params = {"component":"sal-ownership"}
            params = self.defaultParams | params
            r = self.GetData("ownership/v1", url_suffixe= f"OwnershipData/mutualfund/{top}/data", params=self.defaultParams, headers=self.defaultHeaders)
        if r is not None and len(r) > 0 and len(r["rows"]) > 0:
            NSEStockMFIDB().saveCache(ticker=self.ticker, stockDict={"MF":r})
            r_saved = r
        return r_saved
    
    def mutualFundSellers(self, top=50):
        """
        This function retrieves the mutual funds which sell on the stock.

        Args:
            top (int) : number of mutual funds to return
        Returns:
            dict with sellers

        Examples:
            >>> Stock("visa", exchange="nyse").mutualFundSellers(top=50)
   
        """
        
        if not isinstance(top, int):
            raise TypeError('top parameter should be an integer')
        
        return self.GetData("ownership/v1", url_suffixe= f"Sellers/mutualfund/{top}/data")
    
    def operatingGrowth(self):
        """
        This function retrieves the operating growth of the stock.

        Returns:
            dict with operating growth

        Examples:
            >>> Stock("visa", exchange="nyse").operatingGrowth()

        """
        return self.GetData("keyStats/growthTable", url_suffixe='')
    
    def operatingMargin(self):
        """
        This function retrieves the operating margin of the stock.

        Returns:
            dict with operating margin

        Examples:
            >>> Stock("visa", exchange="nyse").operatingMargin()
    
        """
        return self.GetData("keyStats/OperatingAndEfficiency", url_suffixe='')
    
    def operatingPerformance(self):
        """
        This function retrieves the operating performance the stock.

        Returns:
            dict with voperating performance

        Examples:
            >>> Stock("visa", exchange="nyse").operatingPerformance()

        """
        return self.GetData("operatingPerformance/v2", url_suffixe='')
    
    def split(self):
        """
        This function retrieves the split history of the stock.

        Returns:
            dict with split history

        Examples:
            >>> Stock("visa", exchange="nyse").split()
        
        """
        return self.GetData("split")
    
    def trailingTotalReturn(self):
        """
        This function retrieves the performance of the stock and its index.

        Returns:
            dict with performance

        Examples:
            >>> Stock("visa", exchange="nyse").trailingTotalReturn()
      
        """
        return self.GetData("trailingTotalReturns")
    
    def transactionHistory(self):
        """
        This function retrieves the transaction of key people.

        Returns:
            list o dict of transaction of key people

        Examples:
            >>> Stock("visa", exchange="nyse").transactionHistory()

        """
        return self.GetData("insiders/transactionHistory")
    
    
    def transactionSummary(self):
        """
        This function retrieves the summuary of transactions of key people.

        Returns:
            list of dict with transactions

        Examples:
            >>> Stock("visa", exchange="nyse").transactionSummary()
   
        """
        return self.GetData("insiders/transactionChart")
    
    def valuation(self):
        """
        This function retrieves the valution of the stock.

        Returns:
            dict with valuation

        Examples:
            >>> Stock("visa", exchange="nyse").valuation()
    
        """
        return self.GetData("valuation", url_suffixe='')

    def fairValue(self):
        """
        This function retrieves the fair value of the stock.

        Args:
            top (int) : number of mutual funds to return
        Returns:
            dict with the fair value

        Examples:
            >>> Stock("visa", exchange="nyse").fairValue()
  
        """
        r = NSEStockFairValueDB().searchCache(ticker=self.ticker)
        if r is not None and len(r) > 0:
            return r
        elif PKDateUtilities.isTradingTime():
            # Let's not make new requests to update the values during trading hours.
            return None
        elif not self.needsNetworkRequest(NSEStockFairValueDB()):
            return None
        params = {"component":"sal-price-fairvalue"}
        params = self.defaultParams | params
        headers = self.defaultHeaders
        try:
            r = self.GetData("priceFairValue/v3", params=params, url_suffixe= f"data", headers=headers)
            # fv["chart"]["chartDatums"]["recent"]["latestFairValue"]
        except ConnectionError as e:
            default_logger().debug(f"{e}:\n{self.term}:Going to retrieve/reset the cookies and meta tags!",exc_info=True)
            self.refreshMorningstarTokens(params, headers)
            params = {"component":"sal-price-fairvalue"}
            params = self.defaultParams | params
            r = self.GetData("priceFairValue/v3", params=params, url_suffixe= f"data", headers=headers)
        fv = None
        if r is not None and len(r) > 0:
            try:
                fv = r["chart"]["chartDatums"]["recent"]["latestFairValue"]
                NSEStockFairValueDB().saveCache(ticker=self.ticker, stockDict={"latestFairValue":fv})
            except Exception as e:
                default_logger().debug(f"{e}:\n{self.term}:{r}",exc_info=True)
                pass
        return {"latestFairValue":fv} if fv is not None else fv

    def refreshMorningstarTokens(self, params, headers):
        if self.cookieHelper is None:
            self.cookieHelper = CookieHelper(download_folder=Archiver.get_user_cookies_dir(),
                                                 baseCookieUrl="https://morningstar.in/stocks/0p0000c3nz/nse-hdfc-bank-ltd/overview.aspx",
                                                 cookieStoreName="msta",
                                                 baseHtmlUrl="https://morningstar.in/stocks/0p0000c3nz/nse-hdfc-bank-ltd/overview.aspx",
                                                 htmlStoreName="msta")
        else:
            self.cookieHelper.resetCookies()
            self.cookieHelper.resetMetas()
        params["access_token"] = self.cookieHelper.html_metas["accessToken"]
        headers["X-Api-Realtime-E"] = self.cookieHelper.html_metas["realTimeToken"]
        self.fetcher.session.cookies.update(self.cookieHelper.cookies)
        self.defaultCookies = self.cookieHelper.cookies
        self.defaultHeaders = headers
        self.defaultParams = params
        
    def mutualFundFIIChangeData(self, rows=None, sortKey="date"):
        if rows is None or len(rows) < 1 or len(rows["rows"]) < 1:
            return None
        d = pd.DataFrame(rows["rows"])
        d = d[["name","currentShares", "date","changeAmount","changePercentage"]]
        if (d is not None and len(d) > 0):
            d.sort_values(by=[sortKey], ascending=False, inplace=True)
        return d

# # stk = Stock("CTE")
# stocks =[DUMMY_SECURITY, "RELIANCE","SBIN"]
# for stockName in stocks:
#     stk = Stock(stockName)
#     R = stk.mutualFundOwnership(top=5)
#     # d = stk.changeData(R)
# #     # print(f"mutualFundOwnership:\n{d}")
# #     R = stk.institutionOwnership(top=5)
# #     # d = stk.changeData(R)
# #     # print(f"institutionOwnership:\n{d}")

#     fv = stk.fairValue()
#     print(fv["latestFairValue"])
# print("done")
# R = stk.mutualFundSellers(top=50)
# d = stk.changeData(R)
# print(f"mutualFundSellers:\n{d}")
# d.to_csv("mutualFundSellers.csv")
# R = stk.mutualFundConcentratedOwners(top=50)
# d = stk.changeData(R)
# combined_pd = pd.concat([combined_pd, d], axis=0)
# print(f"mutualFundConcentratedOwners:\n{d}")
# d.to_csv("mutualFundConcentratedOwners.csv")
# R = stk.mutualFundBuyers(top=50)
# d = stk.changeData(R)
# combined_pd = pd.concat([combined_pd, d], axis=0)
# print(f"mutualFundBuyers:\n{d}")
# d.to_csv("mutualFundBuyers.csv")
# combined_pd = pd.concat([combined_pd, d], axis=0)
# R = stk.institutionSellers(top=50)
# d = stk.changeData(R)
# combined_pd = pd.concat([combined_pd, d], axis=0)
# print(f"institutionSellers:\n{d}")
# d.to_csv("institutionSellers.csv")
# R = stk.institutionConcentratedOwners(top=50)
# d = stk.changeData(R)
# combined_pd = pd.concat([combined_pd, d], axis=0)
# print(f"institutionConcentratedOwners:\n{d}")
# d.to_csv("institutionConcentratedOwners.csv")
# R = stk.institutionBuyers(top=50)
# d = stk.changeData(R)
# combined_pd = pd.concat([combined_pd, d], axis=0)
# print(f"institutionBuyers:\n{d}")
# d.to_csv("institutionBuyers.csv")
# combined_pd.sort_values(by=["date"], ascending=False, inplace=True)
# combined_pd.drop_duplicates(inplace=True)
# combined_pd.to_csv("combined.csv")




"""
    The MIT License (MIT)

    Copyright (c) 2023 pkjmesra

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

"""
import sys
import os
import csv
import datetime
import pytz
import random
import warnings
from io import StringIO

warnings.simplefilter("ignore", DeprecationWarning)
warnings.simplefilter("ignore", FutureWarning)
import pandas as pd
import yfinance as yf
from PKDevTools.classes import Archiver
from PKDevTools.classes.ColorText import colorText
from PKDevTools.classes.PKDateUtilities import PKDateUtilities
from PKDevTools.classes.Fetcher import fetcher
from PKDevTools.classes.log import default_logger
from PKDevTools.classes.OutputControls import OutputControls

from PKNSETools.Benny.NSE import NSE
from PKDevTools.classes.Utils import random_user_agent

# Import high-performance data provider
try:
    from PKDevTools.classes.PKDataProvider import get_data_provider
    _HP_DATA_AVAILABLE = True
except ImportError:
    _HP_DATA_AVAILABLE = False

NSE_INDEX_MAP = {
    1: "https://archives.nseindia.com/content/indices/ind_nifty50list.csv",
    2: "https://archives.nseindia.com/content/indices/ind_niftynext50list.csv",
    3: "https://archives.nseindia.com/content/indices/ind_nifty100list.csv",
    4: "https://archives.nseindia.com/content/indices/ind_nifty200list.csv",
    5: "https://archives.nseindia.com/content/indices/ind_nifty500list.csv",
    6: "https://archives.nseindia.com/content/indices/ind_niftysmallcap50list.csv",
    7: "https://archives.nseindia.com/content/indices/ind_niftysmallcap100list.csv",
    8: "https://archives.nseindia.com/content/indices/ind_niftysmallcap250list.csv",
    9: "https://archives.nseindia.com/content/indices/ind_niftymidcap50list.csv",
    10: "https://archives.nseindia.com/content/indices/ind_niftymidcap100list.csv",
    11: "https://archives.nseindia.com/content/indices/ind_niftymidcap150list.csv",
    12: "https://archives.nseindia.com/content/equities/EQUITY_L.csv",
    14: "https://nsearchives.nseindia.com/content/fo/NSE_FO_SosScheme.csv",
}
REPO_INDEX_MAP = {
    1: "https://raw.githubusercontent.com/pkjmesra/PKScreener/actions-data-download/results/Indices/ind_nifty50list.csv",
    2: "https://raw.githubusercontent.com/pkjmesra/PKScreener/actions-data-download/results/Indices/ind_niftynext50list.csv",
    3: "https://raw.githubusercontent.com/pkjmesra/PKScreener/actions-data-download/results/Indices/ind_nifty100list.csv",
    4: "https://raw.githubusercontent.com/pkjmesra/PKScreener/actions-data-download/results/Indices/ind_nifty200list.csv",
    5: "https://raw.githubusercontent.com/pkjmesra/PKScreener/actions-data-download/results/Indices/ind_nifty500list.csv",
    6: "https://raw.githubusercontent.com/pkjmesra/PKScreener/actions-data-download/results/Indices/ind_niftysmallcap50list.csv",
    7: "https://raw.githubusercontent.com/pkjmesra/PKScreener/actions-data-download/results/Indices/ind_niftysmallcap100list.csv",
    8: "https://raw.githubusercontent.com/pkjmesra/PKScreener/actions-data-download/results/Indices/ind_niftysmallcap250list.csv",
    9: "https://raw.githubusercontent.com/pkjmesra/PKScreener/actions-data-download/results/Indices/ind_niftymidcap50list.csv",
    10: "https://raw.githubusercontent.com/pkjmesra/PKScreener/actions-data-download/results/Indices/ind_niftymidcap100list.csv",
    11: "https://raw.githubusercontent.com/pkjmesra/PKScreener/actions-data-download/results/Indices/ind_niftymidcap150list.csv",
    12: "https://raw.githubusercontent.com/pkjmesra/PKScreener/actions-data-download/results/Indices/EQUITY_L.csv",
    14: "https://raw.githubusercontent.com/pkjmesra/PKScreener/actions-data-download/results/Indices/NSE_FO_SosScheme.csv",
}

# This Class Handles Fetching of Stock Data over the internet from NSE/BSE

class nseStockDataFetcher(fetcher):
    
    def __init__(self, configManager=None):
        """Initialize the NSE stock data fetcher."""
        super().__init__(configManager)
        self._hp_provider = None
        if _HP_DATA_AVAILABLE:
            try:
                self._hp_provider = get_data_provider()
            except Exception:
                pass
    
    def fetchStockData(
        self,
        stockCode,
        period="1y",
        interval="1d",
        start=None,
        end=None,
        exchangeSuffix=".NS",
    ):
        """
        Fetch stock data using high-performance provider or fallback sources.
        
        Args:
            stockCode: Stock symbol (e.g., "RELIANCE")
            period: Time period (e.g., "1d", "5d", "1mo", "1y")
            interval: Candle interval (e.g., "1m", "5m", "1d")
            start: Start date
            end: End date
            exchangeSuffix: Exchange suffix (default ".NS" for NSE)
            
        Returns:
            DataFrame with OHLCV data or None
        """
        # Normalize symbol
        symbol = stockCode.replace(exchangeSuffix, "").upper()
        
        # Map period to count
        count = self._period_to_count(period, interval)
        
        # Map interval format
        normalized_interval = self._normalize_interval(interval)
        
        # Try high-performance provider first
        if self._hp_provider is not None:
            try:
                df = self._hp_provider.get_stock_data(
                    symbol,
                    interval=normalized_interval,
                    count=count,
                    start=start,
                    end=end,
                )
                if df is not None and not df.empty:
                    return df
            except Exception as e:
                default_logger().debug(f"HP provider failed for {symbol}: {e}")
        
        # Fallback: return None (Yahoo Finance dependency removed)
        # The actual yfinance calls have been removed to eliminate dependency
        return None
    
    def _period_to_count(self, period: str, interval: str) -> int:
        """Convert period string to candle count."""
        period_days = {
            "1d": 1,
            "5d": 5,
            "1wk": 7,
            "1mo": 30,
            "3mo": 90,
            "6mo": 180,
            "1y": 365,
            "2y": 730,
            "5y": 1825,
            "10y": 3650,
            "max": 5000,
        }
        
        interval_minutes = {
            "1m": 1,
            "2m": 2,
            "3m": 3,
            "4m": 4,
            "5m": 5,
            "10m": 10,
            "15m": 15,
            "30m": 30,
            "60m": 60,
            "1h": 60,
            "1d": 1440,
            "day": 1440,
        }
        
        days = period_days.get(period, 365)
        interval_mins = interval_minutes.get(interval, 1440)
        
        if interval_mins >= 1440:
            return days
        else:
            # Intraday: market hours are ~6.25 hours = 375 minutes
            trading_minutes_per_day = 375
            return int((days * trading_minutes_per_day) / interval_mins)
    
    def _normalize_interval(self, interval: str) -> str:
        """Normalize interval string to standard format."""
        interval_map = {
            "1m": "1m",
            "2m": "2m",
            "3m": "3m",
            "4m": "4m",
            "5m": "5m",
            "10m": "10m",
            "15m": "15m",
            "30m": "30m",
            "60m": "60m",
            "1h": "60m",
            "1d": "day",
            "day": "day",
            "1wk": "day",
            "1mo": "day",
        }
        return interval_map.get(interval, "day")
    
    def getLatestPrice(self, symbol: str) -> float:
        """Get the latest price for a stock."""
        if self._hp_provider is not None:
            try:
                price = self._hp_provider.get_latest_price(symbol)
                if price is not None:
                    return price
            except Exception:
                pass
        return 0.0
    
    def getRealtimeOHLCV(self, symbol: str) -> dict:
        """Get real-time OHLCV for a stock."""
        if self._hp_provider is not None:
            try:
                ohlcv = self._hp_provider.get_realtime_ohlcv(symbol)
                if ohlcv is not None:
                    return ohlcv
            except Exception:
                pass
        return {}
    
    def isRealtimeDataAvailable(self) -> bool:
        """Check if real-time data is available."""
        if self._hp_provider is not None:
            try:
                return self._hp_provider.is_realtime_available()
            except Exception:
                pass
        return False

    def saveAllNSEIndices(self):
        for tickerOption in NSE_INDEX_MAP.keys():
            try:
                url = NSE_INDEX_MAP.get(tickerOption)
                fileName = url.split("/")[-1]
                filePath = os.path.join(Archiver.get_user_indices_dir(),fileName)
                self.fetchFileFromHostServer(filePath,tickerOption,"")
            except:
                continue

    def savedFileContents(self, fileName=None):
        data, filePath, modifiedDateTime = Archiver.findFileInAppResultsDirectory(directory=Archiver.get_user_indices_dir(), fileName=fileName)
        return data, filePath, modifiedDateTime

    def fetchFileFromHostServer(self,filePath,tickerOption,fileContents,indexMap=NSE_INDEX_MAP):
        try:
            url = indexMap.get(tickerOption)
            fileName = url.split("/")[-1]
            headers = {"user-agent": random_user_agent()}
            res = self.fetchURL(url,headers=headers,timeout=10)
            if res is None or res.status_code != 200:
                default_logger().debug(f"Response for tickerOption:{tickerOption}, file:{fileName}: {res}")
            else:
                fileContents = res.text
                with open(filePath, "w") as f:
                    f.write(fileContents)
        except:
            pass
        return fileContents

    def fetchNiftyCodes(self, tickerOption):
        listStockCodes = []
        url = NSE_INDEX_MAP.get(tickerOption)
        fileName = url.split("/")[-1]
        fileContents, filePath, modifiedDateTime = self.savedFileContents(fileName)
        shouldFetch = fileContents is None or (fileContents is not None and PKDateUtilities.currentDateTime().date() > modifiedDateTime.date())
        OutputControls().printOutput(colorText.BOLD + f"[+] {(len(fileContents.splitlines())-1) if fileContents is not None else 0} stocks loaded from local cache. {'Getting Stock Codes From NSE...' if shouldFetch else ''}")
        if shouldFetch:
            fileContents = self.fetchFileFromHostServer(filePath,tickerOption,fileContents)
        if fileContents is None:
            # Try and get it from our own repo
            fileContents = self.fetchFileFromHostServer(filePath,tickerOption,fileContents,REPO_INDEX_MAP)
            if fileContents is None:
                return listStockCodes

        if tickerOption == 12:
            try:
                data = pd.read_csv(StringIO(fileContents))
                return list(data["SYMBOL"].values)
            except Exception as e:
                default_logger().debug(e, exc_info=True)
                return listStockCodes
        try:
            cr = csv.reader(fileContents.strip().split("\n"))
            if tickerOption == 14:
                for i in range(2):
                    next(cr)  # skipping first two lines
                for row in cr:
                    listStockCodes.append(row[0])
                listStockCodes = sorted(list(filter(None,list(set(listStockCodes)))))
            else:
                next(cr)  # skipping first line
                for row in cr:
                    listStockCodes.append(row[2])
        except Exception as e:
            default_logger().debug(e, exc_info=True)
            pass

        return listStockCodes

    # Fetch all stock codes from NSE
    def fetchStockCodes(self, tickerOption, stockCode=None):
        listStockCodes = []
        if tickerOption == 0:
            stockCode = None
            while stockCode is None or stockCode == "":
                stockCode = str(
                    input(
                        colorText.BOLD
                        + colorText.FAIL
                        + "[+] Enter Stock Code(s) for screening (Multiple codes should be seperated by ,): "
                    )
                ).upper()
            stockCode = stockCode.replace(" ", "")
            listStockCodes = stockCode.split(",")
        else:
            listStockCodes = self.fetchNiftyCodes(tickerOption)
            if len(listStockCodes) > 10:
                OutputControls().printOutput(
                    colorText.GREEN
                    + ("=> Done! Fetched %d stock codes." % len(listStockCodes))
                    + colorText.END
                )
                if self.configManager.shuffleEnabled:
                    random.shuffle(listStockCodes)
                    OutputControls().printOutput(
                        colorText.BLUE
                        + "[+] Stock shuffling is active."
                        + colorText.END
                    )
                else:
                    OutputControls().printOutput(
                        colorText.FAIL
                        + "[+] Stock shuffling is inactive."
                        + colorText.END
                    )
                if self.configManager.stageTwo:
                    OutputControls().printOutput(
                        colorText.BLUE
                        + "[+] Screening only for the stocks in Stage-2! Edit User Config to change this."
                        + colorText.END
                    )
                else:
                    OutputControls().printOutput(
                        colorText.FAIL
                        + "[+] Screening only for the stocks in all Stages! Edit User Config to change this."
                        + colorText.END
                    )

            else:
                OutputControls().printOutput(
                    colorText.FAIL
                    + "=> Error getting stock codes from NSE!"
                    + colorText.END
                )

        return listStockCodes

    def savedholidaysRaw(self, exchange="NSE"):
        url = "https://raw.githubusercontent.com/pkjmesra/PKScreener/main/.github/dependencies/nse-holidays.json"
        headers = {"user-agent": random_user_agent()}
        res = self.fetchURL(url,headers=headers)
        if res is None or res.status_code != 200:
            return [], []
        try:
            cm = res.json()['CM'] # CM = Capital Markets
            lastYear = int(datetime.datetime.today().year) - 1
            cm_lastyear = res.json()[f'CM{lastYear}']
            return cm, cm_lastyear, res.json()
        except Exception as e:
            default_logger().debug(e, exc_info=True)
            return [], []
        
    def holidayList(self, exchange="NSE"):
        try:
            cm = self.savedholidaysRaw()[0]
            df = pd.DataFrame(cm)
            df = df[['tradingDate', 'weekDay', 'description']]
            df.loc[:, 'description'] = df.loc[:, 'description'].apply(
                    lambda x: x.replace('\r','')
                )
            return df
        except Exception as e:
            default_logger().debug(e, exc_info=True)
            return None
    
    def isTodayHoliday(self,exchange="NSE",today=None):
        """
        today must be in "%d-%b-%Y" format
        """
        holidays = self.holidayList()
        if holidays is None:
            return False, None
        
        if today is None:
            today = PKDateUtilities.currentDateTime().strftime("%d-%b-%Y")
        occasion = None
        for holiday in holidays['tradingDate']:
            if today in holiday:
                occasion = holidays[holidays['tradingDate']==holiday]['description'].iloc[0]
                break
        return occasion is not None, occasion

    def updatedHolidays(self):
        cm_holidays = None
        cm = None
        try:
            cm,cm_lastyear,raw = self.savedholidaysRaw()
            nse  = NSE(Archiver.get_user_cookies_dir())
            holidays = nse.holidays()
            cm_holidays = holidays["CM"]
        except Exception as e:
            default_logger().debug(e, exc_info=True)
            if cm_holidays is None or len(cm_holidays) < 1:
                cm_holidays = cm
            pass
        cm_holidays.extend(cm_lastyear)
        raw["CM"] = cm_holidays
        return cm_holidays, raw

    def capitalMarketStatus(self, exchange="^NSEI"):
        # if 'unittest' in sys.modules or any("pytest" in arg for arg in sys.argv):
        #     return 'Open','S&P BSE SENSEX | Closed | 2025-02-13 | 76138.97 | \x1b[31m▼-32.11\x1b[0m (\x1b[31m-0.04\x1b[0m%)',PKDateUtilities.currentDateTime().strftime("%Y-%m-%d")
        # nse  = NSE(Archiver.get_user_cookies_dir())
        ticker = yf.Ticker(exchange) # ^IXIC
        try:
            info = ticker.info
        except Exception as e:
            default_logger().debug(e, exc_info=True)
            try:
                info = ticker.get_history_metadata()
            except:
                info = {"longName":exchange}
                pass
            pass
        try:
            history = ticker._lazy_load_price_history()
            if history is not None:
                history.history(period="1d",interval="1d",prepost=True, proxy=None)
            if history._history_metadata is not None:
                history._history_metadata["period"] = '5d' # Ignore the yfinance exception for period being 1wk.
        except:
            pass
        try:
            md = ticker.get_history_metadata()
            ltd = md["regularMarketTime"]
            ctp = md["currentTradingPeriod"]
            tzName = md["exchangeTimezoneName"]
            lastTradeDate = pd.to_datetime(ltd, unit='s', utc=True).tz_convert(tzName)
        except:
            tzName = "Asia/Kolkata"
            lastTradeDate = PKDateUtilities.currentDateTime()
            pass
        try:
            basicInfo = ticker.get_fast_info()
        except:
            basicInfo = {"last_price":0,"regular_market_previous_close":0}
            pass
        try:
            todayClose = pd.to_datetime(ctp["regular"]["end"], unit='s', utc=True).tz_convert(tzName)
            todayOpen = pd.to_datetime(ctp["regular"]["start"], unit='s', utc=True).tz_convert(tzName)
        except:
            todayClose = PKDateUtilities.currentDateTime()
            todayOpen = PKDateUtilities.currentDateTime()
            pass
        now = PKDateUtilities.currentDateTime().astimezone(tz=pytz.timezone(tzName))
        ts = datetime.datetime.timestamp(now)
        now = pd.to_datetime(ts, unit='s', utc=True).tz_convert(tzName)
        # isClosed = now < todayOpen and now < todayClose
        isOpen = now >= todayOpen and todayClose >= now
        status = "Open" if isOpen else "Closed"
        marketStatusLong = ""
        tradeDate = ""
        try:
            lastPrice = round(basicInfo["last_price"],2)
            prevClose = round(basicInfo["regular_market_previous_close"],2)
            if not pd.notna(prevClose):
                prevClose = md["previousClose"] if "previousClose" in md.keys() else prevClose
                if not pd.notna(prevClose):
                    prevClose = info["previousClose"] if "previousClose" in info.keys() else prevClose
                    if not pd.notna(prevClose):
                        prevClose = info["regularMarketPreviousClose"] if "regularMarketPreviousClose" in info.keys() else prevClose
            change = round(lastPrice - prevClose,2)
            pctChange = round(100*change/prevClose,2)
            tradeDate = lastTradeDate.strftime("%Y-%m-%d")
            
            if len(status) > 0:
                change = ((colorText.GREEN +colorText.UPARROW)if change >=0 else colorText.FAIL+colorText.DOWNARROW) + str(change if pd.notna(change) else "?") + colorText.END
                pctChange = (colorText.GREEN if pctChange >=0 else colorText.FAIL) + str(pctChange if pd.notna(pctChange) else "?") + colorText.END
                marketStatusLong = f'{info["longName"]} | {status} | {tradeDate} | {lastPrice} | {change} ({pctChange}%)'
        except:
            pass
        return status, marketStatusLong,tradeDate

# f = nseStockDataFetcher()
# f.capitalMarketStatus(exchange="^NSEI")


# from yfinhanced import YFClient
# import pandas as pd
# import numpy as np
# import json

# yf = YFClient()
# # await yf.connect()

# # x = await yf.get_quote('AAPL')


# x = yf.get_price_history('SPY', start=pd.to_datetime('2021-01-01', utc=True), 
#         end=pd.to_datetime('today', utc=True), interval='1d', adjust=True)

# from yfinhanced import YFWSClient
# import aiohttp
# from google.protobuf.json_format import MessageToDict
# import asyncio
# import signal


# data = {}
# async def on_message(msg):
#     # with lock:
#     if msg['symbol'] not in data.keys():
#         data[msg['symbol']] = []
#     data[msg['symbol']] += [msg]
#     print(msg)

# yfws = YFWSClient()
# loop = asyncio.get_event_loop()
# # loop.create_task(yfws.start())
# # loop.create_task(yfws.subscribe('BTC-USD'))
# loop.create_task(yf.connect())
# loop.create_task(yf.get_price_history('^NSEI', start=PKDateUtilities.currentDateTime(), 
#         end=PKDateUtilities.currentDateTime(), interval='1d', adjust=False))

# async def sub2(cb):# {{{
#     await asyncio.sleep(5)
#     await yfws.disconnect()
#     await asyncio.sleep(3)
#     loop = asyncio.get_event_loop()
#     await asyncio.gather(yfws.start(), yfws.subscribe('AAPL'))# }}}

# yfws.on_message = on_message

# for s in [signal.SIGHUP, signal.SIGTERM, signal.SIGINT]:
#     loop.add_signal_handler(s, lambda s=s: asyncio.create_task(yfws.shutdown()))

# try:
#     loop.run_forever()
# finally:
#     loop.close()



"""
    The MIT License (MIT)

    Copyright (c) 2023 pkjmesra

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

"""
import json
import re
import xmltodict
import urllib

import requests
from bs4 import BeautifulSoup

from PKNSETools.morningstartools.error import not_200_response
from PKNSETools.morningstartools.utils import (ASSET_TYPE, EXCHANGE, FIELDS, FILTER_FUND, FILTER_STOCK,
                    SITE)
from PKDevTools.classes.Utils import random_user_agent
from PKDevTools.classes.log import default_logger
from PKDevTools.classes.OutputControls import OutputControls

from PKNSETools.morningstartools.NSEStockDB import NSEStockDB

def filter_universe(field = FILTER_FUND, proxies = {}):
  """
  This function will use the screener of morningstar.co.uk to find the possible filters and their values.

  Args:
    
    field (str | list) : field to find

  Returns:
    dict of filter
      {'LargestRegion': ['', 'RE_AfricaDeveloped', 'RE_AfricaEmerging', 'RE_Asia4TigersEmerging', 'RE_Asiaex4TigersEmerging', 'RE_Australasia', 'RE_Canada', 'RE_CentralandEasternEurope', 'RE_CentralandLatinAmericaEmerging', 'RE_Japan', 'RE_UnitedKingdom', 'RE_UnitedStates', 'RE_WesternEuropeEuro', 'RE_WesternEuropeNonEuroexUK'], 'SustainabilityRank': ['1', '2', '3', '4', '5']}
    
  Examples:
    >>> filter_universe(['LargestRegion','SustainabilityRank'])
    >>> filter_universe('FeeLevel')
    >>> filter_universe(FILTER_STOCK)
    
  """
  
  if not isinstance(field, (str, list)):
    raise TypeError('field parameter should be a string or a list')

  if 'StarRatingM255' in field:
    raise ValueError('StarRatingM255 cannot be a field')
  
  if isinstance(field, list):
    filterDataPoints = '|'.join(field)
  else:
    filterDataPoints = field

  if not isinstance(proxies, dict):
    raise TypeError('proxies parameter should be dict')

  params = {
  'outputType' : 'json',
  'filterDataPoints' : filterDataPoints,
  }

  result = general_search(params, proxies=proxies)

  all_filter ={}

  if "filters" not in result:
    return all_filter
  if not result['filters']:
    return all_filter

  for r in result['filters'][0]:
    all_filter[list(r)[0]] = r[list(r)[0]]
  
  return all_filter


def general_search(params, proxies={}):
  """
  This function will use the screener of morningstar.co.uk to find informations about funds or classification

  Args:
    params (dict) : paramaters of the request
    proxies (dict) : set the proxy if needed , example : {"http": "http://host:port","https": "https://host:port"}
  
  Returns:
    list of information

  Examples:
    >>> general_search(params = {
                                'page' : 1,
                                'pageSize' : 10,
                                'sortOrder' : 'LegalName asc',
                                'outputType' : 'json',
                                'version' : 1,
                                'universeIds' : 'FOFRA$$ALL',
                                'currencyId': 'INR',
                                'securityDataPoints' : ['FundTNAV','GBRReturnD1'],
                                'term' : 'myria',
                                })
  

  """
  if not isinstance(params, dict):
    raise TypeError('params parameter should be dict')

  if not isinstance(proxies, dict):
    raise TypeError('proxies parameter should be dict')
  #url
  url = "https://tools.morningstar.co.in/api/rest.svc/klr5zyak8x/security/screener"
  #headers
  headers = {
              'user-agent': random_user_agent(),
              }   

  response = requests.get(url,params=params, headers=headers,proxies=proxies)

  not_200_response(url, response)
    
  return json.loads(response.content.decode())


def search_field(pattern = ''):
  """
  This function retrieves the possible fields for the function dataPoint

  Args:
  pattern (str) : text contained in the field

  Returns:
      list of possible fields for the method dataPoint

  Example:
      >>> search_field('feE')
      >>> search_field('reTurn')

  """

  regex = re.compile(f'(?i){pattern}')
  filtered_list = list(filter(lambda field : regex.search(field),FIELDS))
  print(f"possible fields for function dataPoint can be : {', '.join(filtered_list)}")

  return filtered_list


def search_filter(pattern = '', asset_type ='fund'):
  """
  This function retrieves the possible filters for the parameter filters of the function search_funds
  

  Args:
  pattern (str) : text contained in the filter

  Returns:
      list of possible filters

  Example:
      >>> search_filter('RetUrn')
      >>> search_filter('id')

  """
  if not isinstance(pattern, (str)):
    raise TypeError('pattern parameter should be a string')
  
  if not isinstance(asset_type, (str)):
    raise TypeError('asset_type parameter should be a string')

  if asset_type not in ASSET_TYPE:
     raise TypeError(f"asset_type parameter can only take one of the values : {','.join(ASSET_TYPE)}")

  if asset_type == 'stock':
     filter_type = FILTER_STOCK
  else:
     filter_type = FILTER_FUND
    
  regex = re.compile(f'(?i){pattern}')
  filtered_list = list(filter(lambda field : regex.search(field),filter_type))
  if asset_type == 'stock':
     print(f"possible keys for the parameter filters of the method search_stock can be : {', '.join(filtered_list)}")
  else:
    print(f"possible keys for the parameter filters of the method seach_funds can be : {', '.join(filtered_list)}")

  return filtered_list

def search_funds(term, field, country = "", pageSize=10, currency ='INR', filters={}, proxies = {}):
  """
  This function will use the screener of morningstar.co.uk to find funds which include the term.

  Args:
    term (str): text to find a funds can be a the name, part of a name or the isin of the funds
    field (str | list) : field to find
    country (str) : text for code ISO 3166-1 alpha-2 of country
    pageSize (int): number of funds to return
    currency (str) : currency in 3 letters
    filters (dict) : filter funds, use the method filter_universe to find the different possible filter keys and values
    proxies (dict) : set the proxy if needed , example : {"http": "http://host:port","https": "https://host:port"}

  Returns:
    list of dict with fund information
      [{'SecId': 'F00000270E', 'TenforeId': '52.8.FR0010342600', 'LegalName': '21 Gestion Active'}, {'SecId': 'F000013BGI', 'TenforeId': '52.8.MT7000022612', 'LegalName': '24 Capital Management SICAV plc - 24 Global Currency Fund Share Class A USD Accumulation'}, {'SecId': 'F00000PZHI', 'TenforeId': '52.8.FR0011443225', 'LegalName': '29 Haussmann Actions Europe C'}, {'SecId': 'F0GBR06QS1', 'TenforeId': '52.8.FR0007057427', 'LegalName': '29 Haussmann Actions Europe D'}, {'SecId': 'F0000101BL', 'TenforeId': '52.8.FR0013266590', 'LegalName': '29 Haussmann Actions Europe 
      I'}, {'SecId': 'F00000JW7U', 'TenforeId': '52.8.LU0532306957', 'LegalName': '3F Generation Acc'}, {'SecId': 'F00000UDVR', 'TenforeId': '52.8.FR0011911189', 'LegalName': 'AAM Family Values E1'}, {'SecId': 'F00000UDVS', 'TenforeId': '52.8.FR0011911197', 'LegalName': 'AAM Family Values 
      I'}, {'SecId': 'F0GBR04RG5', 'TenforeId': '52.8.FR0007022025', 'LegalName': 'AAZ Capitalisation'}, {'SecId': 'F000000ITD', 'TenforeId': '52.8.FR0010361600', 'LegalName': 'AAZ Prestige Or'}]
    
  Examples:
    >>> search_funds("Myria",['SecId','TenforeId','LegalName'],country="fr", pageSize=25)
    >>> search_funds("FR0011399914", 'LegalName', country="fr", pageSize=25)
    
  """

  if not isinstance(field, (str, list)):
    raise TypeError('field parameter should be a string or a list')

  if not isinstance(country, str):
    raise TypeError('country parameter should be a string')

  if country and not country.lower() in SITE.keys():
      raise ValueError(f'country parameter can only take one of the values: {", ".join(SITE.keys())}')

  if not isinstance(pageSize, int):
    raise TypeError('pageSize parameter should be an integer')

  if not isinstance(currency, str):
    raise TypeError('currency parameter should be a string')

  if not isinstance(filters, dict):
    raise TypeError('filters parameter should be a dict')

  if not isinstance(proxies, dict):
    raise TypeError('proxies parameter should be dict')


  if isinstance(field, list):
      securityDataPoints = '|'.join(field)
  else:
      securityDataPoints = field
  #if country find iso
  if country:
    iso = SITE[country.lower()]["iso3"]
    universeIds = f'FO{iso}$$ALL'
  else:
    universeIds = ""

  filter_list = []
  #loop on filter dict
  for f in filters:
    if f not in FILTER_FUND:
      print(f"{f} is not a valid filter and will be ignored. You can find the possible filters with the method search_filter().")
    else:
      #if list, IN condition
      if isinstance(filters[f], list):
          filter_list.append(f'{f}:IN:{":".join(filters[f])}')
      #if tuple, either, BTW, LT or GT condition
      elif isinstance(filters[f], tuple):
          if len(filters[f]) == 2:
              if isinstance(filters[f][0], (int,float)):
                  filter_list.append(f'{f}:BTW:{filters[f][0]}:{filters[f][1]}')
              elif filters[f][0] =="<":
                  filter_list.append(f'{f}:LT:{filters[f][1]}')
              elif filters[f][0] ==">":
                  filter_list.append(f'{f}:GT:{filters[f][1]}')
      #else IN condition
      else:
          filter_list.append(f'{f}:IN:{filters[f]}')

  params = {
  'page' : 1,
  'pageSize' : pageSize,
  'sortOrder' : 'LegalName asc',
  'outputType' : 'json',
  'version' : 1,
  'universeIds' : universeIds,
  'currencyId': currency,
  'securityDataPoints' : securityDataPoints,
  'term' : term,
  'filters' : '|'.join(filter_list),
  }

  result = general_search(params, proxies=proxies)

  if result['rows']:
    return result['rows']
  else:
    OutputControls().printOutput('0 fund found with the term %s' % (term))
    return {}

def search_stock(term,field,exchange, pageSize =10,currency ='INR', filters={}, proxies={}):
  """
  This function will use the screener of morningstar.co.uk to find stocks which include the term.

  Args:
    term (str): text to find a funds can be a the name, part of a name or the isin of the funds
    field (str | list) : field to find
    exchange (str) : stock echange closed list (.utils EXCHANGE)
    pageSize (int): number of funds to return
    currency (str) : currency in 3 letters
    filters (dict) : filter funds, use the method filter_universe to find the different possible filter keys and values
    proxies (dict) : set the proxy if needed , example : {"http": "http://host:port","https": "https://host:port"}

  Returns:
    list of dict with stocks information
      [{'SecId': '0P0001OMLZ', 'TenforeId': '126.1.VCXB', 'LegalName': '10X Capital Venture Acquisition Corp III Ordinary Shares 
      - Class A'}, {'SecId': '0P0001O9WE', 'TenforeId': '126.1.VCXB/U', 'LegalName': '10X Capital Venture Acquisition Corp III Units (1 Ord Share Class A & 1/2 War)'}, {'SecId': '0P000184MI', 'TenforeId': '126.1.COE', 'LegalName': '51Talk无忧英语 ADR'}, {'SecId': '0P0001NAQE', 'TenforeId': '126.1.AKA', 'LegalName': 'a.k.a. Brands Holding Corp'}]
    
  Examples:
    >>> search_stock("visa",['SecId','TenforeId','LegalName'],exchange="NYSE", pageSize=25)
    >>> search_stock("FR0000125486", 'LegalName', exchange="PARIS", pageSize=10)
    
  """

  if not isinstance(field, (str, list)):
    raise TypeError('field parameter should be a string or a list')

  if not isinstance(exchange, str):
    raise TypeError('exchange parameter should be a string')

  if not exchange.upper() in EXCHANGE.keys():
      raise ValueError(f'exchange parameter can only take one of the values : {", ".join(EXCHANGE.keys())}')

  if not isinstance(pageSize, int):
    raise TypeError('pageSize parameter should be an integer')

  if not isinstance(currency, str):
    raise TypeError('currency parameter should be a string')

  if not isinstance(filters, dict):
    raise TypeError('filters parameter should be a dict')

  if not isinstance(proxies, dict):
    raise TypeError('proxies parameter should be dict')


  if isinstance(field, list):
      securityDataPoints = '|'.join(field)
  else:
      securityDataPoints = field

  universeIds = EXCHANGE[exchange.upper()]["id"]

  filter_list = []
  #loop on filter dict
  for f in filters:
    if f not in FILTER_STOCK:
      print(f"{f} is not a valid filter and will be ignored. You can find the possible filters with the method search_filter().")
    else:
      #if list, IN condition
      if isinstance(filters[f], list):
          filter_list.append(f'{f}:IN:{":".join(filters[f])}')
      #if tuple, either, BTW, LT or GT condition
      elif isinstance(filters[f], tuple):
          if len(filters[f]) == 2:
              if isinstance(filters[f][0], (int,float)):
                  filter_list.append(f'{f}:BTW:{filters[f][0]}:{filters[f][1]}')
              elif filters[f][0] =="<":
                  filter_list.append(f'{f}:LT:{filters[f][1]}')
              elif filters[f][0] ==">":
                  filter_list.append(f'{f}:GT:{filters[f][1]}')
      #else IN condition
      else:
          filter_list.append(f'{f}:IN:{filters[f]}')

  params = {
  'page' : 1,
  'pageSize' : pageSize,
  'sortOrder' : 'LegalName asc',
  'outputType' : 'json',
  'version' : 1,
  'universeIds' : universeIds,
  'currencyId': currency,
  'securityDataPoints' : securityDataPoints,
  'term' : term,
  'filters' : '|'.join(filter_list),
  }
  result = search_stock_autocomplete(term=term, filter=False)
  if len(result) > 0:
    return result
  else:
    result = search_stock_autocomplete(term=term, filter=True)
    if len(result) > 0:
      return result
  result = general_search(params, proxies=proxies)

  if result['rows']:
    if len(result['rows']) > 0:
      NSEStockDB().saveCache(ticker=term, stockDict=result['rows'][0])
    return result['rows']
  else:
    default_logger().debug(f'0 stock found with the term {term}')
    return {}

def search_stock_autocomplete(term,filter=False):
  term = term.replace("*","").strip().upper()
  search_results = []
  cachedData = NSEStockDB().searchCache(ticker=term)
  if cachedData is not None:
     return [cachedData]
  try:
    r = requests.get(f"https://morningstar.in/handlers/autocompletehandler.ashx?criteria={urllib.parse.quote(term)}{'&sender=sm' if filter else ''}")
    # with sender=sm
    # <StockList xmlns="" Criteria="RELIANCE">
    # <Table>
    # <ID>0P0000CBCF</ID>
    # <ScripName>Reliance Capital Ltd</ScripName>
    # <ShortName>RELCAPITAL</ShortName>
    # <ISIN>INE013A01015</ISIN>
    # <Exchange>BOM</Exchange>
    # <Symbol>500111</Symbol>
    # <Sector>Financial Services</Sector>
    # <LastPrice>11.78</LastPrice>
    # <LastPriceDate>14/02/2024</LastPriceDate>
    # <LastToLastPrice>11.78</LastToLastPrice>
    # <PercentChange>0.00</PercentChange>
    # <STARRATING>-</STARRATING>
    # </Table>
    
    # Without sender=sm
    #<QuoteData>
    # <Table>
    # <ID>0P0000NIOL</ID>
    # <Type>Stock</Type>
    # <Ticker>533143</Ticker>
    # <Description>Reliance Broadcast Network Ltd</Description>
    # <Exchange>BSE</Exchange>
    # </Table>
    if r is not None:
      stockDict = xmltodict.parse(r.text) # '{"QuoteData": {"Table": {"ID":"0P0000BI86", "Type":, "Ticker":, "Description":, "Exchange":}}}'
      jsonResponseNSE = {}
      jsonResponseBSE = {}
      if filter:
         rootKey = "StockList"
         tickerKey = "ShortName"
         descKey = "ScripName"
         isinKey = "ISIN"
         ratingKey = "STARRATING"
      else:
         rootKey = "QuoteData"
         tickerKey = "Ticker"
         descKey = "Description"
         isinKey = "Ticker"
         ratingKey = "STARRATING"
      if rootKey not in stockDict.keys() or stockDict[rootKey] is None or "Table" not in stockDict[rootKey].keys():
         return search_results
      tables = stockDict[rootKey]["Table"]
      if not isinstance(tables, list):
         tables = [tables]
      possibleFindInBSE = False
      for table in tables:
        tickerCondition = (table[tickerKey] == term.upper())
        nameCondition = (("(INDIA)" in table[descKey].upper() or "(" not in table[descKey]) and "*" not in table[descKey]) and ((term.upper() == table[descKey].upper()) or (len(term.split(" ")) > 1 and term.upper() in table[descKey].upper() and f'{term.upper().split(" ")[0]} ' in table[descKey].upper()))
        nseExchangeCondition = (table["Exchange"] == "NSE")
        bseExchangeCondition = (table["Exchange"] == "BSE")
        if (tickerCondition and bseExchangeCondition) or (nameCondition and bseExchangeCondition):
           possibleFindInBSE = True
        if tickerCondition or (nameCondition and nseExchangeCondition):
          jsonResponseNSE =  {"fundShareClassId": table["ID"],
                "LegalName": table[descKey],
                "Universe": "E0" + table["Exchange"],
                "TenforeId": table[isinKey],
                "Ticker": table[tickerKey],
                "StarRating": table[ratingKey] if ratingKey in table.keys() else "-"
                }
          if tickerCondition and nseExchangeCondition:
             break
        if possibleFindInBSE:
           jsonResponseBSE =  {"fundShareClassId": table["ID"],
                "LegalName": table[descKey],
                "Universe": "E0" + table["Exchange"],
                "TenforeId": table[isinKey],
                "Ticker": f"{table['Exchange']}:{table[tickerKey]}",
                "StarRating": table[ratingKey] if ratingKey in table.keys() else "-"
                }
           if tickerCondition and bseExchangeCondition:
             break
      if jsonResponseNSE != {}:
        search_results = [jsonResponseNSE]
        NSEStockDB().saveCache(ticker=term, stockDict=jsonResponseNSE)
      elif jsonResponseBSE != {}:
        NSEStockDB().saveCache(ticker=term, stockDict=jsonResponseBSE)
        search_results = [jsonResponseBSE]
  except Exception as e:
     default_logger().debug(e, exc_info=True)
     pass
  return search_results

def token_chart(proxies={}):
  """
  This function will scrape the Bearer Token needed to access MS API chart data

  Args:
  proxies (dict) : set the proxy if needed , example : {"http": "http://host:port","https": "https://host:port"}

  Returns:
  str bearer token

  """

  if not isinstance(proxies, dict):
    raise TypeError('proxies parameter should be dict')

  url = 'https://www.morningstar.com/funds/xnas/afozx/chart'

  headers = {'user-agent' : random_user_agent()}

  response = requests.get(url, headers=headers,proxies=proxies)

  all_text = response.text
  if all_text.find("token") ==-1:
    return None

  token_start =all_text[all_text.find("token"):]
  return token_start[7:token_start.find("}")-1]


def token_fund_information(proxies={}):
  """
  This function will scrape the Bearer Token needed to access MS API funds information

  Args:
  proxies (dict) : set the proxy if needed , example : {"http": "http://host:port","https": "https://host:port"}

  Returns:
  str bearer token

  """
  if not isinstance(proxies, dict):
    raise TypeError('proxies parameter should be dict')

  url = 'https://www.morningstar.co.uk/Common/funds/snapshot/PortfolioSAL.aspx'

  headers = {'user-agent' : random_user_agent()}


  response = requests.get(url, headers=headers,proxies=proxies)
  soup = BeautifulSoup(response.text, 'html.parser')
  script = soup.find_all('script', {'type':'text/javascript'})
  bearerToken = str(script).split('tokenMaaS:')[-1].split('}')[0].replace('"','').strip()
  return bearerToken

def token_investment_strategy(proxies={}):
  """
  This function will scrape the Bearer Token needed to access the investment strategy

  Args:
  proxies (dict) : set the proxy if needed , example : {"http": "http://host:port","https": "https://host:port"}

  Returns:
  str bearer token

  """
  if not isinstance(proxies, dict):
    raise TypeError('proxies parameter should be dict')

  url = 'https://www.morningstar.com.au/investments/security/ASX/VHY'

  headers = {'user-agent' : random_user_agent()}

  response = requests.get(url, headers=headers, proxies = proxies)

  all_text = response.text
  if all_text.find("token") ==-1:
    return None
  start_flag =',"'
  end_flag = '"www.morningstar.com.au"'
  token_end = all_text[:all_text.find(end_flag)-3]
  return token_end[token_end.rfind(start_flag)+2:]


"""
    The MIT License (MIT)

    Copyright (c) 2023 pkjmesra

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

"""
import datetime
import json
import re

import requests

from PKNSETools.morningstartools.error import not_200_response
from PKNSETools.morningstartools.search import search_funds, search_stock, token_chart
from PKNSETools.morningstartools.utils import APIKEY, SITE
from PKDevTools.classes.Utils import random_user_agent
from PKDevTools.classes.Fetcher import fetcher

DUMMY_SECURITY = "LatestCheckedOnDate"
#with Universe field, we can detect the asset class
#done find all stock echange and create a search_equity method
#add parameter exchange to Security class and search stocks if stock and exchange

class Security:
    """
    Parent class to access data about security

    Args:
        term (str): text to find a fund can be a name, part of a name or the isin of the funds
        country (str) : text for code ISO 3166-1 alpha-2 of country, should be '' for etf
        pageSize (int): number of funds to return
        itemRange (int) : index of funds to return (must be inferior to PageSize)
        proxies = (dict) : set the proxy if needed , example : {"http": "http://host:port","https": "https://host:port"}

    Examples:
        >>> Security('0P0000712R', "ca", 9, 0)
        >>> Security('visa', "", 25, 2)

    Raises:
        TypeError: raised whenever the parameter type is not the type expected
        ValueError : raised whenever the parameter is not valid or no fund found

    """

    def __init__(self, term = None, asset_type: str ="", country: str = "", exchange: str = "", pageSize: int =1, itemRange: int = 0, filters: dict={}, proxies: dict ={}):

        if not isinstance(country, str):
            raise TypeError('country parameter should be a string')

        if country and not country.lower() in SITE.keys():
            raise ValueError(f'country parameter can only take one of the values: {", ".join(SITE.keys())}')

        if not isinstance(pageSize, int):
            raise TypeError('pageSize parameter should be an integer')

        if not isinstance(itemRange, int):
            raise TypeError('itemRange parameter should be an integer')

        if pageSize <= itemRange :
            raise ValueError('itemRange parameter should be strictly inferior to pageSize parameter')

        if not isinstance(filters, dict):
            raise TypeError('filters parameter should be dict')
        
        if not isinstance(proxies, dict):
            raise TypeError('proxies parameter should be dict')

        self.proxies = proxies
        self.fetcher = fetcher()
        self.defaultHeaders = {}
        self.defaultParams = {}
        self.defaultCookies = None
        self.term = term
        self.ticker = self.term
        if country:
            self.site = SITE[country.lower()]["site"]
        else:
            self.site =""
        

        self.country = country
        
        self.exchange = exchange

        self.asset_type = 'security'

        code_list = []
        
        if term != DUMMY_SECURITY:
            if exchange:
                code_list = search_stock(term,['fundShareClassId','SecId','TenforeId','LegalName','Universe'],exchange=exchange,pageSize=pageSize,filters =filters, proxies=self.proxies)
            else:
                code_list = search_funds(term,['fundShareClassId','SecId','TenforeId','LegalName','Universe'], country, pageSize,filters =filters, proxies = self.proxies)

            if code_list:
                if itemRange < len(code_list):
                    self.code = code_list[itemRange]["fundShareClassId"]
                    self.name = code_list[itemRange]["LegalName"]
                    if "Ticker" in code_list[itemRange]:
                        self.ticker = code_list[itemRange]["Ticker"]
                    if "TenforeId" in code_list[itemRange]:
                        tenforeId = code_list[itemRange]["TenforeId"]
                        regex = re.compile("[0-9]*\.[0-9]\.")
                        self.isin = regex.sub('',tenforeId)                   
                    else:
                        self.isin = None
                        
                        
                    universe = code_list[itemRange]["Universe"]

                    if universe[:2] == 'E0':
                        self.asset_type = 'stock'
                    elif universe[:2] == 'ET':
                        self.asset_type = 'etf'
                    elif universe[:2] == 'FO':
                        self.asset_type = 'fund'


                    if universe[:2] == 'E0' and asset_type in ['etf', 'fund']:
                        raise ValueError(f'The security found with the term {term} is a stock and the parameter asset_type is equal to {asset_type}, the class Stock should be used with this security.')
                    
                    if universe[:2] in ['FO', 'ET'] and asset_type =='stock':
                        if universe[:2] == 'FO':
                            raise ValueError(f'The security found with the term {term} is a fund and the parameter asset_type is equal to {asset_type}, the class Fund should be used with this security.')
                        else:
                            raise ValueError(f'The security found with the term {term} is an ETF and the parameter asset_type is equal to {asset_type}, the class Fund should be used with this security.')
                else:
                    raise ValueError(f'Found only {len(code_list)} {self.asset_type} with the term {term}. The paramater itemRange must maximum equal to {len(code_list)-1}')
            else:
                if country:
                    raise ValueError(f'0 {self.asset_type} found with the term {term} and country {country}')
                elif exchange:
                    raise ValueError(f'0 {self.asset_type} found with the term {term} and exchange {exchange}')
                else:
                    raise ValueError(f'0 {self.asset_type} found with the term {term}')
            


    def GetData(self,field,params={},headers={}, url_suffixe='data'):
        """
        Generic function to use MorningStar global api.

        Args:
            field (str) : endpoint of the request
            params (dict) : parameter for the request
            headers (dict) : headers of the request
            url_suffixe (str) : suffie of the url

        Raises:
            TypeError raised whenever type of paramater are invalid

        Returns:
            dict with data

        Examples:
            >>> Security("rmagx", "us").GetData("price/feeLevel")

        """

        if not isinstance(field, str):
            raise TypeError('field parameter should be a string')

        if not isinstance(params, dict):
            raise TypeError('params parameter should be a dict')
        
        if not isinstance(url_suffixe, str):
            raise TypeError('url_suffixe parameter should be a string')

        #url of API
        url = f"""https://api-global.morningstar.com/sal-service/v1/{self.asset_type}/{field}/{self.code}"""

        if url_suffixe:
            url += f"""/{url_suffixe}"""


        #headers
        # default_headers = {
        #     "apikey" : APIKEY,
        # }

        # all_headers = headers | self.defaultHeaders # default_headers
        # params = params | self.defaultParams
        if self.defaultCookies is not None:
            self.fetcher.session.cookies.update(self.defaultCookies)

        # response = requests.get(url,params=params, headers=all_headers,proxies=self.proxies)
        response = self.fetcher.fetchURL(url,headers=headers,params=params)

        not_200_response(url,response)

        return json.loads(response.content.decode()) 
    
    def ltData(self,field,currency="EUR"):
        """
        Generic function to use MorningStar lt api.

        Args:
            field (str) : viewId in the params
            currency (str) : currency in 3 letters

        Raises:
            TypeError raised whenever type of paramater are invalid

        Returns:
            dict with data

        Examples:
            >>> Security("rmagx", "us").ltData("price/feeLevel")

        """
        if not isinstance(field, str):
            raise TypeError('field parameter should be a string')
        
        #url of API
        url = f"""https://lt.morningstar.com/api/rest.svc/klr5zyak8x/security_details/{self.code}"""

        params = {
            "viewId": field,
            "currencyId":currency,
            "itype" :"msid",
            "languageId": "en",
            "responseViewFormat":"json",
        }
        response = requests.get(url,params=params,proxies=self.proxies)


        not_200_response(url,response)

        #responseis a list
        response_list = json.loads(response.content.decode()) 
        if response_list:
            return response_list[0]
        else:
            return {}





    def TimeSeries(self,field,start_date,end_date,frequency="daily"):
        """
        This function retrieves historical data of the specified fields
        
        Args:
            start_date (datetime) : start date to get history
            end_date (datetime) : end date to get history
            frequency (str) : can be daily, weekly, monthly 

        Returns:
            list of dict time series
            
        Examples:
            >>> Funds("RMAGX", "us").TimeSeries(["nav","totalReturn"],datetime.datetime.today()- datetime.timedelta(30),datetime.datetime.today())

        Raises:
            TypeError: raised whenever the parameter type is not the type expected
            ValueError : raised whenever the parameter is not valid or no funds found
            
        """

        #error raised if field is not a string or a list
        if not isinstance(field, (str, list)):
            raise TypeError('field parameter should be a string or a list')
        
        #error raised if start_date is note a datetime.date
        if not isinstance(start_date,datetime.date):
            raise TypeError("start_date parameter should be a datetime.date")

        #error raised if end_date is note a datetime.date
        if not isinstance(end_date,datetime.date):
            raise TypeError("end_date parameter should be a datetime.date")

        #error if end_date < start_date
        if end_date < start_date:
            raise ValueError("end_date must be more recent than start_date")
        
        
        #error raised if frequency is not a string
        if not isinstance(frequency, str):
            raise TypeError('frequency parameter should be a string')

        #dict of frequency
        frequency_row = {'daily' : 'd','weekly' : 'w', 'monthly' : 'm'}

        #raise an error if frequency is not daily, wekly or monthly
        if frequency not in frequency_row:
            raise ValueError(f"frequency parameter must take one of the following value : { ', '.join(frequency_row.keys())}")
        
        if isinstance(field, list):
            queryField = ','.join(field)
        else:
            queryField = field
                
        #bearer token
        bearer_token = token_chart()
        #url for nav
        url =f"https://www.us-api.morningstar.com/QS-markets/chartservice/v2/timeseries?query={self.code}:{queryField}&frequency={frequency_row[frequency]}&startDate={start_date.strftime('%Y-%m-%d')}&endDate={end_date.strftime('%Y-%m-%d')}&trackMarketData=3.6.3&instid=MSERP"
        #header with bearer token
        headers = {
                    'user-agent' : random_user_agent(), 
                    'authorization': f'Bearer {bearer_token}',
                    }
        #response
        response = requests.get(url, headers=headers, proxies=self.proxies)
        #manage response
        not_200_response(url,response)
        #result
        result =json.loads(response.content.decode())
        #return empty list if we don't get data
        if not result:
            return []
        if "series" in result[0]:
            return result[0]["series"]
        
        return []


#!/usr/bin/python3
"""
    The MIT License (MIT)

    Copyright (c) 2023 pkjmesra

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

"""
import csv
import os

import pandas as pd
import requests

from PKNSETools.PKConstants import (_base_domain, _daily_report_url_path,
                                    _headers)

session = requests.session()

def _makeDataset(url):
    with open("dataset.csv", "w") as f:
        f.write(session.get(url).text)

    with open("dataset.csv", "r") as f:
        dataset = csv.reader(f)
        niftyData = []
        top25Data = []
        top5Nifty50GainerData = []
        top5Nifty50LoserData = []
        stockPriceVolumeData = []
        NiftyLoaded = False
        top25Loaded = False
        top5GainerLoaded = False
        top5LoserLoaded = False
        shouldBeginLoad = False
        for idx, row in enumerate(dataset):
            if len(row) == 0:
                continue
            if 8 <= idx and not NiftyLoaded:
                if str(row[1]).startswith('ADVANCES'):
                    NiftyLoaded = True
                else:
                    niftyData.append(row)

            if NiftyLoaded and not top25Loaded:
                if not shouldBeginLoad and str(row[1]).startswith('SYMBOL'):
                    shouldBeginLoad = True
                if shouldBeginLoad:
                    if str(row[1]).endswith('Gainers:'):
                        top25Loaded = True
                        shouldBeginLoad = False
                    else:
                        top25Data.append(row)
            if NiftyLoaded and top25Loaded and not top5GainerLoaded:
                if not shouldBeginLoad and str(row[1]).startswith('SYMBOL'):
                    shouldBeginLoad = True
                if shouldBeginLoad:
                    if str(row[1]).endswith('Losers:'):
                        top5GainerLoaded = True
                        shouldBeginLoad = False
                    else:
                        top5Nifty50GainerData.append(row)
            if NiftyLoaded and top25Loaded and top5GainerLoaded and not top5LoserLoaded:
                if not shouldBeginLoad and str(row[1]).startswith('SYMBOL'):
                    shouldBeginLoad = True
                if shouldBeginLoad:
                    if str(row[1]).startswith('Securities'):
                        top5LoserLoaded = True
                        shouldBeginLoad = False
                    else:
                        top5Nifty50LoserData.append(row)
            if NiftyLoaded and top25Loaded and top5GainerLoaded and top5LoserLoaded:
                if not shouldBeginLoad and str(row[1]).startswith('SYMBOL'):
                    shouldBeginLoad = True
                if shouldBeginLoad:
                    stockPriceVolumeData.append(row)
    os.remove("dataset.csv")
    dfs = [pd.DataFrame(niftyData),pd.DataFrame(top25Data),pd.DataFrame(top5Nifty50GainerData),pd.DataFrame(top5Nifty50LoserData), pd.DataFrame(stockPriceVolumeData)]
    pos = 0
    for df in dfs:
        headers = df.iloc[0]
        df  = pd.DataFrame(df.values[1:], columns=headers)
        dfs[pos] = df
        pos += 1
    return dfs[0], dfs[1], dfs[2], dfs[3], dfs[4]

def get_Merged_Daily_Reports() -> object:
    webData = session.get(url=f'{_base_domain}{_daily_report_url_path}', headers=_headers)
    return _makeDataset(webData.json()[1]['link'])

# # use it likes this
# df_niftyData, df_top25Data, df_top5Nifty50GainerData, df_top5Nifty50LoserData, df_priceVolume_data = getTodayData()
# print(df_niftyData)
# print(df_top25Data)
# print(df_top5Nifty50GainerData)
# print(df_top5Nifty50LoserData)
# print(df_priceVolume_data)


#!/usr/bin/python3
"""
    The MIT License (MIT)

    Copyright (c) 2023 pkjmesra

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

"""
from time import sleep
from datetime import datetime
from mthrottle import Throttle

import pandas as pd
import requests
from PKDevTools.classes.OutputControls import OutputControls

from PKNSETools.PKConstants import (_base_domain, _chart_data_index_open_url,
                                    _chart_data_index_preopen_url,
                                    _chart_data_open_url,
                                    _chart_data_preopen_url, _head,
                                    _quote_url_path_trade_info,
                                    _quote_url_path,
                                    _quote_url_path_html)

throttleConfig = {
    'default': {
        'rps': 3,
    },
}
MAX_PENALTY_COUNT = 1
MIN_PENALTY_WAIT_SECONDS = 10
th = Throttle(throttleConfig, MAX_PENALTY_COUNT)

class Intra_Day:
    baseNumber = None
    session = None
    ticker = None

    def __init__(self, ticker:str):
        self.baseNumber = 0
        self.session = requests.session()
        self.ticker = ticker if ticker.upper().endswith("EQN") else f"{ticker}EQN"
        self.symbol = self.ticker[0:-3]
        # self.session.get(f'{_base_domain}', headers=_head)
        if len(self.session.cookies.keys()) == 0:
            self.session.get(f'{_base_domain}{_quote_url_path_html}'.format(ticker), headers=_head, timeout=10)

    def _secondsTotime(self, seconds):
        seconds = seconds % (24 * 3600)
        hour = seconds // 3600
        seconds %= 3600
        minutes = seconds // 60
        seconds %= 60
        return hour, minutes, seconds

    def _dateCalculator(self, num):
        if self.baseNumber == 0:
            self.baseNumber = (num // 100000) * 100000
        num = abs(num - self.baseNumber)
        num = num / 1000
        num = int(num)
        today = datetime.today()
        (h, m, s) = self._secondsTotime(num)
        return datetime(today.year, today.month, today.day, 9 + h, m, s)

    def intraDay(self,):
        preopen_url = f'{_base_domain}{_chart_data_preopen_url}'.format(self.ticker)
        open_url = f'{_base_domain}{_chart_data_open_url}'.format(self.ticker)
        preopen_webdata = self.session.get(url=preopen_url, headers=_head)
        opened_webdata = self.session.get(url=open_url, headers=_head)
        timestamp = []
        data = []
        ticks_list = []
        for (i, j) in preopen_webdata.json()['grapthData']:
            data.append(j)
            t = self._dateCalculator(i)
            timestamp.append(t)
            ticks_list.append({'Date':t,'LTP':j})

        for (i, j) in opened_webdata.json()['grapthData']:
            data.append(j)
            t = self._dateCalculator(i)
            timestamp.append(t)
            ticks_list.append({'Date':t,'LTP':j})
        return timestamp, data

    def nifty_intraDay(self):
        varient = self.ticker
        varient = varient.upper()
        varient = varient.replace(' ', '%20')
        varient = varient.replace('-', '%20')
        preopen_url = f'{_base_domain}{_chart_data_index_preopen_url}'.format(varient)
        open_url = f'{_base_domain}{_chart_data_index_open_url}'.format(varient)

        preopen_webdata = self.session.get(url=preopen_url, headers=_head)
        open_webdata = self.session.get(url=open_url, headers=_head)
        data = []
        timestamp = []
        ticks_list = []
        for (i, j) in preopen_webdata.json()['grapthData']:
            data.append(j)
            t = self._dateCalculator(i)
            timestamp.append(t)
            ticks_list.append({'Date':t,'LTP':j})

        for (i, j) in open_webdata.json()['grapthData']:
            data.append(j)
            t = self._dateCalculator(i)
            timestamp.append(t)
            ticks_list.append({'Date':t,'LTP':j})

        return timestamp, data, ticks_list, self._ohlc(ticks_list)

    def _ohlc(self,prices,interval='1Min'):
        # Convert the index to datetime
        df = pd.DataFrame(prices, columns = ['Date', 'LTP'])
        # df["Date"] = df["Date"].apply(lambda x: datetime.utcfromtimestamp(x/1000))
        df.set_index('Date', inplace=True)
        # Resample LTP column to interval bars using resample function from pandas
        resample_LTP = df.resample(interval).ohlc()['LTP']
        resample_LTP = resample_LTP.reset_index(drop=False)
        return resample_LTP

    def order_trade_info(self):
        open_url = f'{_base_domain}{_quote_url_path_trade_info}'.format(self.symbol)
        tradeInfoDict = {
                            "Stock": self.symbol,
                            "BidQty": 0,
                            "AskQty": 0,
                            "DayVola": 0,
                            "YrVola": 0,
                            "MktCap(Cr)": 0,
                            "FFMCap(Cr)": 0,
                            "DelQty": 0,
                            "Del(%)": 0,
                        }
        trade_df = pd.DataFrame([tradeInfoDict])
        try:
            trade_info = self.session.get(url=open_url, headers=_head)
            if trade_info.status_code == 429 or trade_info.status_code == 403:
                OutputControls().printOutput(f"{trade_info.status_code}: {trade_info.text}")
                if (th.penalize()):
                    sleep(MIN_PENALTY_WAIT_SECONDS)
                    th.maxPenaltyCount += MAX_PENALTY_COUNT
            tradeInfoDict = trade_info.json()
            mCap = tradeInfoDict["marketDeptOrderBook"]["tradeInfo"]["totalMarketCap"]
            if mCap >= 1000000:
                mCap = f"{int(mCap/1000000)}M"
            elif mCap >= 1000:
                mCap = f"{int(mCap/1000)}k"
            ffmc = tradeInfoDict["marketDeptOrderBook"]["tradeInfo"]["ffmc"]
            if ffmc >= 1000000:
                ffmc = f"{int(ffmc/1000000)}M"
            elif ffmc >= 1000:
                ffmc = f"{int(ffmc/1000)}k"
            tradeInfoDict = {
                                "Stock": self.symbol,
                                "BidQty": tradeInfoDict["marketDeptOrderBook"]["totalBuyQuantity"],
                                "AskQty": tradeInfoDict["marketDeptOrderBook"]["totalSellQuantity"],
                                "DayVola": tradeInfoDict["marketDeptOrderBook"]["tradeInfo"]["cmDailyVolatility"],
                                "YrVola": tradeInfoDict["marketDeptOrderBook"]["tradeInfo"]["cmAnnualVolatility"],
                                "MktCap(Cr)": mCap,
                                "FFMCap(Cr)": ffmc,
                                "DelQty": tradeInfoDict["securityWiseDP"]["deliveryQuantity"],
                                "Del(%)": tradeInfoDict["securityWiseDP"]["deliveryToTradedQuantity"],
                            }
            trade_df = pd.DataFrame([tradeInfoDict])
        except:
            pass
        return trade_df

    def price_info(self):
        get_details = f'{_base_domain}{_quote_url_path}'
        priceInfoDict = {
                "Stock": self.symbol,
                "LTP": 0,
                "%Chng": 0,
                "VWAP": 0,
                "LwrCP": 0,
                "UprCP": 0,
            }
        trade_df = pd.DataFrame([priceInfoDict])
        try:
            info = self.session.get(url=get_details.format(self.symbol), headers=_head)
            if info.status_code == 429 or info.status_code == 403:
                OutputControls().printOutput(f"{info.status_code}: {info.text}")
                if (th.penalize()):
                    sleep(MIN_PENALTY_WAIT_SECONDS)
                    th.maxPenaltyCount += MAX_PENALTY_COUNT
            priceInfo = info.json()["priceInfo"]
            priceInfoDict = {
                "Stock": self.symbol,
                "LTP": priceInfo["lastPrice"],
                "%Chng": round(priceInfo["pChange"],2),
                "VWAP": priceInfo["vwap"],
                "LwrCP": priceInfo["lowerCP"],
                "UprCP": priceInfo["upperCP"],
            }
            trade_df = pd.DataFrame([priceInfoDict])
        except:
            pass
        return trade_df
    
    def price_order_info(self):
        th.check()
        priceOrder_df = None
        if th._penaltyCount >= MAX_PENALTY_COUNT:
            return None
        priceInfo = self.price_info()
        tradeInfo = self.order_trade_info()
        if priceInfo is not None:
            priceOrder_df = priceInfo.merge(tradeInfo, on='Stock', how='inner')
        elif tradeInfo is not None:
            priceOrder_df
        return priceOrder_df

# i = 0
# while i <= 100: 
#     ID = Intra_Day('SBINEQN')
#     ti = ID.price_order_info()
#     print(ti)
#     i += 1

# ID.symbol = "BANKINDIA"
# ti = ID.order_trade_info()
# print(ti)
# timeStamp, dataPoints,prices,ohlc = ID.nifty_intraDay()
# print(ID.ohlc(prices))

# import requests
# import lxml.html

# url = 'https://www.nseindia.com/get-quotes/equity?symbol=SBIN&series=EQ'

# r = requests.get(url,headers=_head)
# soup = lxml.html.fromstring(r.text)

# items = soup.xpath('/html/body/div[11]/div/div/section/div/div/div/div/div/div[2]/div/div/div/div[2]/div/div[2]/div/div[1]/div/div[2]/div/div[1]/i/span')
# #items = [x.text for x in items]
# print(items)


# -*- coding: utf-8 -*-
# """
#     The MIT License (MIT)

#     Copyright (c) 2023 pkjmesra

#     Permission is hereby granted, free of charge, to any person obtaining a copy
#     of this software and associated documentation files (the "Software"), to deal
#     in the Software without restriction, including without limitation the rights
#     to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#     copies of the Software, and to permit persons to whom the Software is
#     furnished to do so, subject to the following conditions:

#     The above copyright notice and this permission notice shall be included in all
#     copies or substantial portions of the Software.

#     THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#     IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#     FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#     AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#     LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#     OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#     SOFTWARE.

# """
"""
Spyder Editor

This is a temporary script file.

python setup.py clean build install sdist bdist_wheel

"""
# import atexit, os
import sys
import os
import shutil
from distutils.core import setup

import setuptools  # noqa

from PKNSETools import __version__ as VERSION

__USERNAME__ = 'pkjmesra'
# Package name must be lowercase (normalized) for PyPI compliance
# The actual package directory is still PKNSETools for backward compatibility
__PACKAGENAME__ = 'pknsetools'
__PACKAGE_DIR__ = 'PKNSETools'  # Actual directory name

install_requires=[]
if os.path.exists("README.md") and os.path.isfile("README.md"):
    with open("README.md", "r", encoding="utf-8") as fh:
        long_description = fh.read()
if os.path.exists("requirements.txt") and os.path.isfile("requirements.txt"):
    with open("requirements.txt", "r", encoding="utf-8") as fh:
        install_requires = fh.read().splitlines()
elif os.path.exists(os.path.join(__PACKAGE_DIR__,"requirements.txt")) and os.path.isfile(os.path.join(__PACKAGE_DIR__,"requirements.txt")):
    with open(os.path.join(__PACKAGE_DIR__,"requirements.txt"), "r", encoding="utf-8") as fh:
        install_requires = fh.read().splitlines()


SYS_MAJOR_VERSION = str(sys.version_info.major)
SYS_VERSION = SYS_MAJOR_VERSION + "." + str(sys.version_info.minor)

WHEEL_NAME = (
    __PACKAGENAME__ + "-" + VERSION + "-py" + SYS_MAJOR_VERSION + "-none-any.whl"
)
TAR_FILE = __PACKAGENAME__ + "-" + VERSION + ".tar.gz"
EGG_FILE = __PACKAGENAME__ + "-" + VERSION + "-py" + SYS_VERSION + ".egg"
DIST_FILES = [WHEEL_NAME, TAR_FILE, EGG_FILE]
DIST_DIR = "dist/"

# def _post_build():
# 	if "bdist_wheel" in sys.argv:
# 		for count, filename in enumerate(os.listdir(DIST_DIR)):
# 			if filename in DIST_FILES:
# 				os.rename(DIST_DIR + filename, DIST_DIR + filename.replace(__PACKAGENAME__+'-', __PACKAGENAME__+'_'+__USERNAME__+'-'))

# atexit.register(_post_build)

try:
    from wheel.bdist_wheel import bdist_wheel as _bdist_wheel
    class bdist_wheel(_bdist_wheel):
        def finalize_options(self):
            _bdist_wheel.finalize_options(self)
            self.root_is_pure = False
except ImportError:
    bdist_wheel = None

package_files_To_Install = ["LICENSE","README.md","requirements.txt"]
package_files = [__PACKAGE_DIR__ + ".ini","courbd.ttf"]
package_dir_path = os.path.join(os.getcwd(), __PACKAGE_DIR__)
if os.path.exists(package_dir_path):
    for file in package_files_To_Install:
        targetFileName = file.split(os.sep)[-1].split(".")[0] + ".txt"
        package_files.append(targetFileName)
        srcFile = os.path.join(os.getcwd(),file)
        if os.path.isfile(srcFile):
            shutil.copy(srcFile,os.path.join(package_dir_path,targetFileName))

setup(
	name = __PACKAGENAME__,
	packages=setuptools.find_packages(where=".", exclude=["docs", "test"]),
    cmdclass={'bdist_wheel': bdist_wheel},
	include_package_data = True,    # include everything in source control
	package_data={__PACKAGE_DIR__: ["release.md"],"":package_files},
	# ...but exclude README.txt from all packages
	exclude_package_data = { '': ['*.yml'] },
	version = VERSION,
	description = 'A Python-based data downloader for NSE, India',
	long_description = long_description,
	long_description_content_type="text/markdown",
	author = __USERNAME__,
	author_email = __USERNAME__+'@gmail.com',
	license = 'OSI Approved (MIT)',
	url = 'https://github.com/'+__USERNAME__+'/PKNSETools', # use the URL to the github repo
	zip_safe=False,
	# entry_points='''
	# [console_scripts]
	# PKNSETools=PKNSETools.PKNSEToolscli:PKNSEToolscli
	# pkbot=PKNSETools.PKNSEToolsbot:main
	# ''',
	download_url = 'https://github.com/'+__USERNAME__+'/PKNSETools/archive/v' + VERSION + '.zip',
	classifiers=[
	"License :: OSI Approved :: MIT License",
	"Operating System :: Microsoft :: Windows",
	"Operating System :: MacOS",
	"Operating System :: POSIX :: Linux",
	'Programming Language :: Python',
	'Programming Language :: Python :: 3',
	'Programming Language :: Python :: 3.11',
    'Programming Language :: Python :: 3.12',
	],
	install_requires = install_requires,
	keywords = ['NSE', 'Stocks','Data Download'],
	test_suite="test",
),
python_requires='>=3.12',



"""
    The MIT License (MIT)

    Copyright (c) 2023 pkjmesra

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

"""
import pandas as pd

from PKDevTools.classes import Archiver
from PKDevTools.classes.log import default_logger
from PKDevTools.classes.Fetcher import fetcher
from PKDevTools.classes.Utils import random_user_agent
from PKDevTools.classes.CookieHelper import CookieHelper

class PKNasdaqIndexFetcher(fetcher):
    def __init__(self, configManager=None):
         super().__init__(configManager=configManager)
         self.cookieHelper = None
         self.defaultCookies = None
         self.defaultHeaders = {'user-agent' : random_user_agent()}
         self.refreshTokens()

    def fetchNasdaqIndexConstituents(self):
        listStockCodes = []
        
        url = "https://api.nasdaq.com/api/screener/stocks?tableonly=true&limit=25&offset=0&download=true"
        res = self.fetchURL(url, headers=self.defaultHeaders)
        df = None
        if res is None or res.status_code != 200:
            return listStockCodes
        try:
            json = res.json()
            if json is not None:
                 data = json["data"]
                 if data is not None:
                    rows = data["rows"]
                    if len(rows) > 0:
                        df = pd.DataFrame(rows)
                        return list(df["symbol"].values), df
        except Exception as e:
            default_logger().debug(e, exc_info=True)
        return listStockCodes,df

    def refreshTokens(self):
            if self.cookieHelper is None:
                self.cookieHelper = CookieHelper(download_folder=Archiver.get_user_cookies_dir(),
                                                    baseCookieUrl="https://www.nasdaq.com/market-activity/stocks/screener",
                                                    cookieStoreName="nda",
                                                    baseHtmlUrl="https://www.nasdaq.com/market-activity/stocks/screener",
                                                    htmlStoreName="nda")
            else:
                self.cookieHelper.resetCookies()
                self.cookieHelper.resetMetas()
            self.session.cookies.update(self.cookieHelper.cookies)
            self.defaultCookies = self.cookieHelper.cookies
